/*
 * Argus Software
 * Copyright (c) 2000-2007 QoSient, LLC
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/* 
 * $Id: //depot/argus/argus-3.0/clients/include/argus_parser.h#49 $
 * $DateTime: 2006/05/25 02:07:01 $
 * $Change: 858 $
 */


#ifndef ArgusParser_h
#define ArgusParser_h

#include <argus_def.h>
#include <argus_out.h>
#include <net/nff.h>

#if defined(ARGUS_THREADS)
#include <pthread.h>
#endif

#define ARGUS_PRINTSRC	0x0001
#define ARGUS_PRINTDST	0x0002
#define ARGUS_ACTTIME	0x0004
#define ARGUS_IDLETIME	0x0008
#define ARGUS_DIST	0x0010

#define ARGUS_PRINTGMT	0x0020
#define ARGUS_PRINTNET	0x0022
 
#define ARGUS_ENCODE_ASCII       0
#define ARGUS_ENCODE_64          1
#define ARGUS_ENCODE_32          2
#define ARGUS_HEXDUMP            3

#define ARGUS_MAX_S_OPTIONS      134

#define RA_PRINTPROTO		1
#define RA_PRINTSRCID		2
 
#define RAMON_TOPN		1
#define RAMON_MATRIX		2
#define RAMON_SVCS		3
#define RAMON_SINGLE		4

#define RA_FIXED_WIDTH		1
#define RA_VARIABLE_WIDTH	2

#define RA_SINGLE_QUOTED	'\''
#define RA_DOUBLE_QUOTED	'\"'

#define ARGUS_MAX_REMOTE		256

#define RABINS_HASHTABLESIZE	0x100


/* the ArgusRecordStruct (ns) is a single point data structure
   for clients to use to process and report on ARGUS flow data.  
   To support this rather nebulous function, the ns  provides
   a canonical record buffer which supports a completely
   parsed and formatted record (exploded view), and indexes
   to the various sections of the ns structure.

   A queue header struct is supported so that the strucutre
   can be placed in a queue, and there is an hstruct pointer
   provided so that if the record is hashed, the hash entry
   can be found quickly, ie for removal.

   Because ns data is time series data, there is a desire to
   support an array style data structure to provide the
   ability to  maintain some aspects fo the time-series
   qualities.  This struct was enabled in earlier versions
   using the 'H' (histogram) option.  In order to support the
   concept of the Stream Block Processor, however, this data
   structure needs to be embeeded in the ns itself.

   The concept is that the ns itself holds the aggregate
   stats for the time series array.  Each  member is an
   ns, which can further subdivided into additional time series
   structures.  For most applications, this structure will
   be NULL, indicating that subdivision is not being done.

*/

 
struct ArgusHashStruct {
   unsigned int len, hash;
   unsigned int *buf; 
}; 
 
struct ArgusHashTableHdr {
   struct ArgusHashTableHdr *nxt, *prv;
   struct ArgusHashTable *htbl;
   struct ArgusHashStruct hstruct;
   void *object;
};


struct ArgusHashTable {
   unsigned int size;
   struct ArgusHashTableHdr **array;
#if defined(ARGUS_THREADS)
   pthread_mutex_t lock;
#endif /* ARGUS_THREADS */
};

struct ArgusAdjustStruct {
   int mode, turns, modify, soft, zero;
   int value, count, qual, slen, len;
   float tperiod;
   float spkts, sbytes, sappbytes;
   float scpkts, scbytes, scappbytes;
   float dpkts, dbytes, dappbytes;
   float dcpkts, dcbytes, dcappbytes;
   struct timeval start, end;
   struct tm RaStartTmStruct, RaEndTmStruct;
   int startSecs, size;
   char *filename, *filterstr;
   struct nff_program filter;
};

struct RaBinStruct {
   int status, value, size, timeout;
   struct timeval stime, etime;
   struct ArgusQueueStruct *queue;
   struct ArgusHashTable *htable;
   unsigned char ArgusSrcDataMask[16],ArgusDstDataMask[16];
};

struct RaBinProcessStruct {
   int status, start, end, size;
   struct timeval startpt, endpt, rtime;
   int arraylen, len, max, count, index;
   int scalesecs;
   struct RaBinStruct **array;
   struct ArgusAdjustStruct nadp;
};

#define ARGUS_NSR_STICKY		0x00100000
 
struct ArgusRecordStruct {
   struct ArgusQueueHeader qhdr;
   unsigned int status, dsrindex, trans, timeout, idle;
   struct RaBinProcessStruct *bins;
   struct ArgusHashTableHdr *htblhdr;
   struct ArgusQueueStruct *nsq;
   struct ArgusRecordHeader hdr;
   struct ArgusDSRHeader *dsrs[ARGUSMAXDSRTYPE];
   struct ArgusCanonRecord canon;
   float srate, drate, sload, dload, dur, avgdur;
   long long offset;
};

struct ArgusCIDRAddr {
   u_char type, len, masklen, pad; 
   u_int addr[4], mask[4];
   char *str;
};

#define ARGUS_MAX_PRINT_ALG	114
#define MAX_PRINT_ALG_TYPES	114

struct ArgusParserStruct {
   int status, RaParseCompleting, RaParseDone;
   int RaTasksToDo, ArgusReliableConnection;
   char *ArgusProgramName, *RaTimeFormat, *RaTimeZone;
   char *ArgusProgramArgs, *ArgusProgramOptions;
   struct timeval ArgusRealTime, ArgusGlobalTime;
   struct timeval RaClientTimeout, RaClientUpdate;
   struct timeval RaStartTime, RaEndTime;
   struct tm RaStartFilter, RaLastFilter;
   struct tm *RaTmStruct;

#if defined(ARGUS_THREADS)
   pthread_t thread;
   pthread_mutex_t lock;
   pthread_t remote, output, dns;
#endif /* ARGUS_THREADS */

   int ArgusSessionId, NonBlockingDNS;

   char *ArgusPidFile, *ArgusPidPath;

   struct ArgusRecordStruct *ns;
   struct ArgusRecordStruct argus;
   struct ArgusCanonRecord canon;

   struct ArgusOutputStruct *ArgusOutput;
   struct ArgusListStruct *ArgusOutputList;
   struct ArgusListStruct *ArgusNameList;

   struct ArgusInput *ArgusRemoteFDs[ARGUS_MAX_REMOTE];
   int ArgusActiveFDs;
   int ArgusMaxListen;
   int ArgusLfd;
   int ArgusAdjustTime;
   int ArgusConnectTime;
   int ArgusReverse;
   int ArgusPrintXml;

   char *RadiumArchive;

   unsigned int ArgusID, ArgusIDType;

   struct timeval ArgusReportTime;
   struct timeval ArgusUpdateInterval;
   struct timeval ArgusMarReportInterval;

   long long ArgusTotalRecords;
   long long ArgusTotalMarRecords;
   long long ArgusTotalFarRecords;
   long long ArgusTotalPkts, ArgusTotalSrcPkts, ArgusTotalDstPkts;
   long long ArgusTotalBytes, ArgusTotalSrcBytes, ArgusTotalDstBytes;

   signed char aflag, Aflag, bflag, Bflag;
   signed char cflag, Cflag, dflag, Dflag, eflag, Eflag;
   signed char fflag, Fflag, gflag, Gflag, Hflag;
   signed char idflag, jflag, lflag, Lflag, mflag;
   signed char notNetflag, Oflag, pflag, Pflag, qflag, Qflag;
   signed char Netflag, nflag, Normflag, Pctflag, pidflag;

   signed char tflag, uflag, Wflag, vflag, Vflag, iflag;
   signed char Iflag, Tflag, rflag, Rflag, Sflag, xflag;
   signed char Xflag, XMLflag, zflag, Zflag;

   int debugflag, RaInitialized;

   char RaFieldDelimiter, RaFieldQuoted; 
   signed char RaFieldWidth, RaWriteOut;

   float Pauseflag;

   int hfield, prfield, pfield, Uflag;
   int sNflag, eNflag;
   int startime_t, lasttime_t, timeout;
   char *estr, *Hstr, *Mflag, *sflag;
   char RaLabelStr[0x10000], *RaLabel;
   int ArgusRandomSeed;
   int RaLabelCounter;

   int RaSOptionIndex;
   char *RaSOptionStrings[ARGUS_MAX_S_OPTIONS];

   int ArgusFilterFiledes[2];
   int ArgusControlFiledes[2];

   int RaCumulativeMerge;
   int RaFlowMajorModified;
   int RaAllocHashTableHeaders;
   int RaAllocArgusRecord;
   int RaThisActiveIndex;
   int RaThisFlowNum;
   int RaThisModelNum;
   int RaParseError;
   int ArgusMinuteUpdate;
   int ArgusHourlyUpdate;
   int RaHistoMetricSeries;
   int RaHistoMetricLog;

   double RaHistoLogInterval;
   double RaHistoBinSize;
   double RaHistoStart, RaHistoStartLog;
   double RaHistoEnd, RaHistoEndLog;
   struct ArgusRecordStruct **RaHistoRecords;

   int ArgusPortNum;
   int RaHistoBins;
   int RaCloseInputFd;

   int RaPollMode;
   int RaAgMode;
   int RaMonMode;
   int RaZeroMode;
   int RaPrintMode;
   int RaCursesMode;
   int RaPrintIndex;
   int RaPrintStartTime;
   int RaPrintLastTime;
   int RaExplicitDate;
   int RaWildCardDate;

   char *RaFlowModelFile, *exceptfile;
   char *projstr, *dbustr, *dbpstr;
   char *ustr, *pstr, *timearg, *wfile;

   char *ArgusFlowModelFile;

   char *ArgusLocalFilter;
   char *ArgusRemoteFilter;
   char *ArgusDisplayFilter;

   char *ArgusBindAddr;

   struct nff_program ArgusFilterCode;
   struct nff_program ArgusDisplayCode;

   struct RaFlowModelStruct *RaFlowModel;

   struct ArgusCIDRAddr ArgusCIDRBuffer, *ArgusCIDRPtr;

   char ArgusOriginalBuffer[MAXARGUSRECORD];
   struct ArgusRecord ArgusInitCon;
   struct ArgusRecord *ArgusOriginal;

   struct ArgusModeStruct *ArgusModeList;
   struct ArgusModeStruct *ArgusMaskList;
   struct ArgusInput *ArgusCurrentInput;
   struct ArgusInput *ArgusInputFileList;
   struct ArgusInput *ArgusRemoteHostList;
   struct ArgusListStruct *ArgusWfileList;

   void (*RaPrintAlgorithm)(char *, struct ArgusRecordStruct *);
   void (*RaPrintAlgorithms[ARGUS_MAX_PRINT_ALG])(struct ArgusParserStruct *, char *, struct ArgusRecordStruct *);

   char RaDebugString[MAXSTRLEN];

   char ArgusSrcUserData[0x10000];
   char ArgusDstUserData[0x10000];
};

#ifdef ArgusParse
struct ArgusParserStruct *ArgusNewParser(char *);
void ArgusInitializeParser(struct ArgusParserStruct *);

#else
extern struct ArgusParserStruct *ArgusNewParser(char *);
extern void ArgusInitializeParser(struct ArgusParserStruct *);
#endif

#endif
