/*
 * Argus Software
 * Copyright (c) 2000-2007 QoSient, LLC
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

/* 
 * $Id: //depot/argus/argus-3.0/clients/ratop/ratop.c#73 $
 * $DateTime: 2006/05/24 09:49:59 $
 * $Change: 856 $
 */

/*
 * ratop.c  - top program for argus data.  Integrates with the ntais.
 * 
*/

#define ARGUS_RECORD_MODIFIED	0x0100

#include <unistd.h>
#include <stdlib.h>
#include <sys/ioctl.h>

#include <argus_util.h>
#include <argus_client.h>
#include <argus_main.h>
#include <argus_filter.h>

#include <rabins.h>
#include <rasplit.h>
#include <ratop.h>

#include <signal.h>
#include <ctype.h>

#include <argus_sort.h>
#include <argus_cluster.h>

#include <glob.h>

struct timeval RaStartTime = {0x7FFFFFFF, 0x7FFFFFFF};
struct timeval RaEndTime = {0, 0};

#define ARGUS_REMOTE_FILTER	1
#define ARGUS_LOCAL_FILTER	2
#define ARGUS_DISPLAY_FILTER	3


#define RAMON_NETS_CLASSA	0
#define RAMON_NETS_CLASSB	1
#define RAMON_NETS_CLASSC	2
#define RAMON_NETS_CLASS	3

#define RA_DIRTYBINS            0x20

extern void ArgusInitAggregatorStructs(struct ArgusAggregatorStruct *);

void RaTopLoop (struct ArgusParserStruct *);
void RaRefreshDisplay(struct ArgusParserStruct *);
void RaOutputModifyScreen (void);
void RaOutputHelpScreen (void);

struct RaBinProcessStruct *RaBinProcess = NULL;

struct RaTopProcessStruct *RaTopNewProcess(struct ArgusParserStruct *parser);

#define RATOPSTARTINGINDEX	2

struct RaTopProcessStruct {
   int status, timeout; 
   int value, size;
   struct ArgusRecordStruct *ns;
   struct ArgusQueueStruct *queue, *delqueue;
   struct ArgusHashTable *htable;
   struct nff_program filter;
};

struct RaTopProcessStruct *RaTopProcess = NULL;

float RaUpdateRate = 1.0;
int RaTopRealTime = 0;
int RaCursorOffset = 0;
int RaCursorX = 0;
int RaCursorY = 0;

struct timeval ArgusLastRealTime = {0, 0};
struct timeval ArgusLastTime     = {0, 0};
struct timeval ArgusThisTime     = {0, 0};
struct timeval ArgusCurrentTime  = {0, 0};

struct timeval dLastTime = {0, 0};
struct timeval dRealTime = {0, 0};
struct timeval dThisTime = {0, 0};
struct timeval dTime     = {0, 0};

long long thisUsec = 0;
long long lastUsec = 0;

struct ArgusQueueStruct *ArgusModelerQueue;
struct ArgusQueueStruct *ArgusFileQueue;
struct ArgusQueueStruct *ArgusProbeQueue;

void RaResizeHandler (int);

void
ArgusClientInit (struct ArgusParserStruct *parser)
{
   struct ArgusAdjustStruct *nadp = NULL;
   struct ArgusInput *input = NULL;
   struct ArgusModeStruct *mode; 
   char outputfile[MAXSTRLEN];
   int i = 0, size = 1;

   outputfile[0] = '\0';
   parser->RaWriteOut = 1;

   if (!(parser->RaInitialized)) {
      (void) signal (SIGHUP,  (void (*)(int)) RaParseComplete);
      (void) signal (SIGTERM, (void (*)(int)) RaParseComplete);
      (void) signal (SIGQUIT, (void (*)(int)) RaParseComplete);
      (void) signal (SIGINT,  (void (*)(int)) RaParseComplete);
      (void) signal (SIGWINCH,(void (*)(int)) RaResizeHandler);
      (void) signal (SIGPIPE, SIG_IGN);

      parser->RaClientTimeout.tv_sec  = 0;
      parser->RaClientTimeout.tv_usec = 10000;

#if defined(ARGUS_CURSES)
      RaInitCurses(parser);
#endif
      parser->RaInitialized++;

      if (parser->nflag > 1)
         parser->pfield = 5;

      parser->NonBlockingDNS = 1;
      parser->RaCumulativeMerge = 1;
      parser->timeout = 60;

      if (parser->ArgusInputFileList != NULL) {
         parser->RaTasksToDo = 1;
         if ((input = parser->ArgusRemoteHostList) == NULL)
            parser->timeout = 0;
      }

      if ((input = parser->ArgusRemoteHostList) != NULL) 
         parser->RaTasksToDo = 1;

      if (parser->ArgusFlowModelFile) {
         if ((ArgusAggregator = ArgusParseAggregator(parser, NULL)) == NULL)
            ArgusLog (LOG_ERR, "ArgusClientInit: ArgusParseAggregator error");

      } else
         if ((ArgusAggregator = ArgusNewAggregator(parser, NULL)) == NULL)
            ArgusLog (LOG_ERR, "ArgusClientInit: ArgusNewAggregator error");

      if (parser->Hstr != NULL)
         ArgusHistoMetricParse(parser, ArgusAggregator);

      if ((ArgusModelerQueue = ArgusNewQueue()) == NULL)
         ArgusLog(LOG_ERR, "ArgusClientInit: RaNewQueue error %s", strerror(errno));

      if ((ArgusProbeQueue = ArgusNewQueue()) == NULL)
         ArgusLog(LOG_ERR, "ArgusClientInit: RaNewQueue error %s", strerror(errno));

      if ((ArgusFileQueue = ArgusNewQueue()) == NULL)
         ArgusLog(LOG_ERR, "ArgusClientInit: RaNewQueue error %s", strerror(errno));

      if ((RaTopProcess = RaTopNewProcess(parser)) == NULL)
         ArgusLog (LOG_ERR, "ArgusClientInit: RaTopNewProcess error");

      if (parser->vflag) 
         ArgusReverseSortDir++;
 
      if ((ArgusSorter = ArgusNewSorter()) == NULL)
         ArgusLog (LOG_ERR, "ArgusClientInit: ArgusNewSorter error %s", strerror(errno));

      ArgusSorter->ArgusSortAlgorithms[0] = ArgusSortAlgorithmTable[ARGUSSORTPKTSCOUNT];

      if ((RaBinProcess = (struct RaBinProcessStruct *)ArgusCalloc(1, sizeof(*RaBinProcess))) == NULL)
         ArgusLog (LOG_ERR, "ArgusClientInit: ArgusCalloc error %s", strerror(errno));

      nadp = &RaBinProcess->nadp;

      nadp->mode   = -1;
      nadp->modify =  0;
      nadp->slen   =  2;
 
      if (parser->aflag)
         nadp->slen = parser->aflag;

      if ((mode = parser->ArgusModeList) != NULL) {
         int i, x, ind;
         while (mode) {
            for (i = 0, ind = -1; i < ARGUSSPLITMODENUM; i++) {
               if (!(strncasecmp (mode->mode, RaSplitModes[i], strlen(RaSplitModes[i])))) {
                  ind = i;
                  break;
               }
            }

            if (ind >= 0) {
               char *mptr = NULL;
               switch (ind) {
                  case ARGUSSPLITRATE:  {   /* "%d:%d[yMwdhms]" */
                     struct ArgusModeStruct *tmode = NULL; 
                     nadp->mode = ind;
                     if ((tmode = mode->nxt) != NULL) {
                        mptr = tmode->mode;
                        if (isdigit((int)*tmode->mode)) {
                           char *ptr = NULL;
                           nadp->count = strtol(tmode->mode, (char **)&ptr, 10);
                           if (*ptr++ != ':') 
                              usage();
                           tmode->mode = ptr;
                        }
                     }
                  }

                  case ARGUSSPLITTIME: /* "%d[yMwdhms] */
                     nadp->mode = ind;
                     if ((mode = mode->nxt) != NULL) {
                        if (isdigit((int)*mode->mode)) {
                           char *ptr = NULL;
                           nadp->value = strtol(mode->mode, (char **)&ptr, 10);
                           if (ptr == mode->mode)
                              usage();
                           else {
                              switch (*ptr) {
                                 case 'y':
                                    nadp->qual = ARGUSSPLITYEAR;  
                                    size = nadp->value * 31556926;
                                    break;
                                 case 'M':
                                    nadp->qual = ARGUSSPLITMONTH; 
                                    size = nadp->value * 2629744;
                                    break;
                                 case 'w':
                                    nadp->qual = ARGUSSPLITWEEK;  
                                    size = nadp->value * 604800;
                                    break;
                                 case 'd':
                                    nadp->qual = ARGUSSPLITDAY;   
                                    size = nadp->value * 86400;
                                    break;
                                 case 'h':
                                    nadp->qual = ARGUSSPLITHOUR;  
                                    size = nadp->value * 3600;
                                    break;
                                 case 'm':
                                    nadp->qual = ARGUSSPLITMINUTE;
                                    size = nadp->value * 60;
                                    break;
                                  default:
                                    nadp->qual = ARGUSSPLITSECOND;
                                    size = nadp->value;
                                    break;
                              }
                           }
                        }
                        if (mptr != NULL)
                            mode->mode = mptr;
                     }

                     nadp->modify = 1;

                     if (ind == ARGUSSPLITRATE) {
                        /* need to set the flow idle timeout value to be equal to or
                           just a bit bigger than (nadp->count * size) */

                        ArgusParser->timeout = (nadp->count * size);
                     }

                     ArgusSorter->ArgusSortAlgorithms[0] = ArgusSortStartTime;
                     ArgusSorter->ArgusSortAlgorithms[1] = NULL;
                     break;

                  case ARGUSSPLITSIZE:
                  case ARGUSSPLITCOUNT:
                     nadp->mode = ind;
                     nadp->count = 1;

                     if ((mode = mode->nxt) != NULL) {
                        if (isdigit((int)*mode->mode)) {
                           char *ptr = NULL;
                           nadp->value = strtol(mode->mode, (char **)&ptr, 10);
                           if (ptr == mode->mode)
                              usage();
                           else {
                              switch (*ptr) {
                                 case 'B':   
                                 case 'b':  nadp->value *= 1000000000; break;
                                  
                                 case 'M':   
                                 case 'm':  nadp->value *= 1000000; break;
                                  
                                 case 'K':   
                                 case 'k':  nadp->value *= 1000; break;
                              }
                           }
                        }
                     }
                     ArgusSorter->ArgusSortAlgorithms[0] = NULL;
                     break;

                  case ARGUSSPLITNOMODIFY:
                     nadp->modify = 0;
                     break;

                  case ARGUSSPLITSOFT:
                     nadp->soft++;
                     break;

                  case ARGUSSPLITZERO:
                     nadp->zero++;
                     break;
               }

            } else {
               if (!(strncasecmp (mode->mode, "nomerge", 7))) {
                  parser->RaCumulativeMerge = 0;
               } else
               if (!(strncasecmp (mode->mode, "merge", 5))) {
                  parser->RaCumulativeMerge = 1;
               } else
               if (!(strncasecmp (mode->mode, "rtime", 5)) ||
                  (!(strncasecmp (mode->mode, "realtime", 8)))) {
                  char *ptr = NULL;
                  RaTopRealTime++;
                  if ((ptr = strchr(mode->mode, ':')) != NULL) {
                     double value = 0.0;
                     char *endptr = NULL;
                     ptr++;
                     value = strtod(ptr, &endptr);
                     if (ptr != endptr) {
                        RaUpdateRate = value;
                     }
                  }

               } else {
                  for (x = 0, i = 0; x < MAX_SORT_ALG_TYPES; x++) {
                     if (!strncmp (ArgusSortKeyWords[x], mode->mode, strlen(ArgusSortKeyWords[x]))) {
                        ArgusSorter->ArgusSortAlgorithms[i++] = ArgusSortAlgorithmTable[x];
                        break;
                     }
                  }
               }
            }

            mode = mode->nxt;
         }
      }

      RaBinProcess->size = size;

      if (nadp->mode < 0) {
         nadp->mode = ARGUSSPLITCOUNT;
         nadp->value = 10000;
         nadp->count = 1;
      }

      /* if content substitution, either time or any field, is used,
         size and count modes will not work properly.  If using
         the default count, set the value so that we generate only
         one filename.

         if no substitution, then we need to add "aa" suffix to the
         output file for count and size modes.
      */

      if ((parser->ArgusWfileList != NULL) && (!(ArgusListEmpty(parser->ArgusWfileList)))) {
         struct ArgusWfileStruct *wfile = NULL;
         int count = parser->ArgusWfileList->count;

         if (count > 1)
            usage();

         if ((wfile = (struct ArgusWfileStruct *) ArgusPopFrontList(parser->ArgusWfileList, ARGUS_NOLOCK)) != NULL) {
            strncpy (outputfile, wfile->filename, MAXSTRLEN);
 
            if ((strchr(outputfile, '%')) || (strchr(outputfile, '$'))) {
               switch (nadp->mode) {
                  case ARGUSSPLITCOUNT:
                     nadp->count = -1;
                     break;

                  case ARGUSSPLITSIZE:
                     for (i = 0; i < nadp->slen; i++) 
                        strcat(outputfile, "a");
                     break;
               }

            } else {
               switch (nadp->mode) {
                  case ARGUSSPLITSIZE:
                  case ARGUSSPLITCOUNT:
                     for (i = 0; i < nadp->slen; i++) 
                        strcat(outputfile, "a");
                     break;
               }
            }

            if (!(strchr(outputfile, '%'))) {
               switch (nadp->mode) {
                  case ARGUSSPLITTIME:
                    break;
               }
            }

            nadp->filename = strdup(outputfile);
            setArgusWfile (parser, outputfile, NULL);
         }
      }

      if ((input = parser->ArgusRemoteHostList) != NULL) {
         while (input) {
            if ((input->fd = ArgusGetServerSocket (input, 5)) >= 0) {
               if ((ArgusReadConnection (parser, input, ARGUS_SOCKET)) >= 0) {
                  ArgusParser->ArgusRemoteFDs[ArgusParser->ArgusActiveFDs++] = input;
               } else {
                  sprintf (parser->RaDebugString, "ArgusReadConnection '%s': %s", input->hostname, strerror(errno));
               }
            } else {
               sprintf (parser->RaDebugString, "ArgusGetServerSocket '%s': %s", input->hostname, strerror(errno));
            }
            input = input->nxt;
         }
      }

      RaTopLoop (parser);
   }
}


#define ARGUS_FILE_LIST_PROCESSED	0x1000

void
RaTopLoop (struct ArgusParserStruct *parser)
{
   int retn = 0, fd = -1;

   parser->RaParseDone = 0;
   RaRefreshDisplay(parser);

   while (1) {
      if (parser->RaTasksToDo) {
         struct ArgusInput *input = NULL;

         ArgusInitializeParser(parser);
         RaTopStartTime.tv_sec  = 0;
         RaTopStartTime.tv_usec = 0;
         RaTopStopTime.tv_sec   = 0;
         RaTopStopTime.tv_usec  = 0;

         if ((!(parser->status & ARGUS_FILE_LIST_PROCESSED)) && ((input = parser->ArgusInputFileList) != NULL)) {
            while (input) {
               if (*input->filename != '-') {
                  if ((input->file = fopen(input->filename, "r")) >= 0) {
                     sprintf (parser->RaDebugString, "RaTopLoop() Processing Input File %s", input->filename);
                     if (((ArgusReadConnection (parser, input, ARGUS_FILE)) >= 0)) {
                        parser->ArgusCurrentInput = input;

                        if (input->ostart != -1) {
                           input->offset = input->ostart;
                           if (fseek(input->file, input->offset, SEEK_SET) >= 0)
                              ArgusReadFileStream(parser);
                        } else
                           ArgusReadFileStream(parser);

                        if (input->file != NULL) {
                           fclose(input->file);
                           input->file = NULL;
                        }
                        sprintf (parser->RaDebugString, "RaTopLoop() Processing Input File %s done.", input->filename);

                     } else {
                        sprintf (parser->RaDebugString, "ArgusReadConnection '%s': %s", input->hostname, strerror(errno));
                     }

                  } else {
                     sprintf (parser->RaDebugString, "open '%s': %s", input->filename, strerror(errno));
                  }
               }
               input = input->nxt;
            }
            parser->status |= ARGUS_FILE_LIST_PROCESSED;
         }

         if (parser->ArgusActiveFDs) {
            sprintf (parser->RaDebugString, "RaTopLoop() Processing Remote Inputs.");
            ArgusReadStream(parser);
         }

         parser->RaTasksToDo = 0;

      } else {
         RaRefreshDisplay(parser);
         usleep(50000);
      }

      ArgusClientTimeout ();
   }

   if (fd >= 0) {
      ArgusShutDown (0);
   } else {
      ArgusWindowClose();
      retn = 1;
   }

#if defined(ARGUSDEBUG)
   ArgusDebug (1, "main () exiting with %d\n", retn);
#endif

   exit (retn);
}

void RaArgusInputComplete (struct ArgusInput *input) {
   RaWindowImmediate = TRUE;
   RaRefreshDisplay(ArgusParser);
}

void
RaParseComplete (int sig)
{
   if (sig >= 0) {
      ArgusWindowClose();

      if (sig == SIGINT) {
         ArgusShutDown(0);
         exit(0);
      }
   }
}

void
RaResizeHandler (int sig)
{
#ifdef ARGUS_CURSES
   struct winsize size;
   RaScreenResize = TRUE;

   if (ioctl(fileno(stdout), TIOCGWINSZ, &size) == 0) {
#if defined(__FreeBSD__) || (__NetBSD__) || (__OpenBSD__)
      resizeterm(size.ws_row, size.ws_col);
#else
#if defined(HAVE_SOLARIS)
#else
      resize_term(size.ws_row, size.ws_col);
#endif
#endif
      wrefresh(curscr);   /* Linux needs this */
   }
#endif

#ifdef ARGUSDEBUG 
   ArgusDebug (1, "ArgusResizeHandler(%d)\n", sig);
#endif
}


char *ArgusGenerateProgramArgs(struct ArgusParserStruct *);
char RaProgramArgs[MAXSTRLEN];

char *
ArgusGenerateProgramArgs(struct ArgusParserStruct *parser)
{
   char *retn = RaProgramArgs;
   struct ArgusModeStruct *mode = NULL;
   struct ArgusInput *input = NULL;
   int i;
   
   sprintf (retn, "%s ", parser->ArgusProgramName);

   if (parser->ArgusActiveFDs) {
      if (parser->Sflag)
         sprintf (&retn[strlen(retn)], "-S ");
      else
         sprintf (&retn[strlen(retn)], "-r ");

      for (i = 0; i < parser->ArgusActiveFDs; i++) {
         if ((input = parser->ArgusRemoteFDs[i]) != NULL) {
            if (parser->Sflag)
               sprintf (&retn[strlen(retn)], "%s:%d ", input->hostname, input->portnum);
            else
               sprintf (&retn[strlen(retn)], "%s ", input->filename);
         }
      }

   } else {
      if (RaProject && RaTable) {
         sprintf (&retn[strlen(retn)], "-P %s:%s ", RaProject, RaTable);
      }
   }

   if ((mode = parser->ArgusModeList) != NULL) { 
      sprintf (&retn[strlen(retn)], "-M ");
      while (mode) { 
         sprintf (&retn[strlen(retn)], "%s ", mode->mode);
         mode = mode->nxt;
      }
   }

   if (((mode = parser->ArgusMaskList) != NULL) || (ArgusAggregator->mask == 0)) {
      sprintf (&retn[strlen(retn)], "-m ");
      while (mode) {
         sprintf (&retn[strlen(retn)], "%s ", mode->mode);
         mode = mode->nxt;
      }
   }

   if (parser->Hstr)
      sprintf (&retn[strlen(retn)], "-H %s ", parser->Hstr);

   if ((parser->ArgusDisplayFilter) || parser->ArgusLocalFilter || parser->ArgusRemoteFilter) {
      sprintf (&retn[strlen(retn)], "- ");
      if (parser->ArgusDisplayFilter)
         sprintf (&retn[strlen(retn)], "display '%s' ", parser->ArgusDisplayFilter);
      if (parser->ArgusLocalFilter)
         sprintf (&retn[strlen(retn)], "local '%s' ", parser->ArgusLocalFilter);
      if (parser->ArgusRemoteFilter) 
         sprintf (&retn[strlen(retn)], "remote '%s' ", parser->ArgusRemoteFilter);
   }
   return (retn);
}


void RaTopSortQueue (struct ArgusSorterStruct *, struct ArgusQueueStruct *);
 
void
RaTopSortQueue (struct ArgusSorterStruct *sorter, struct ArgusQueueStruct *queue)
{
   struct ArgusQueueHeader *qhdr = queue->start;
   struct nff_insn *fcode = NULL;
   int i = 0, x = 0, cnt = queue->count;
 
   if (queue->array != NULL) {
      ArgusFree(queue->array);
      queue->array = NULL;
   } 

   if (cnt > 0) {
      fcode = sorter->filter.bf_insns;
      if ((queue->array = (struct ArgusQueueHeader **) ArgusCalloc(1, sizeof(struct ArgusQueueHeader *) * (cnt + 1))) != NULL) {
         for (i = 0; i < cnt; i++) {
            int keep = 0;
            if (fcode) {
               if (ArgusFilterRecord (fcode, (struct ArgusRecordStruct *)qhdr) != 0)
                  keep = 1;
            } else
               keep = 1;
      
            if (keep)
               queue->array[x++] = qhdr;
            qhdr = qhdr->nxt;
         }
         queue->array[i] = NULL;
         qsort ((char *) queue->array, x, sizeof (struct ArgusQueueHeader *), ArgusSortRoutine);

      } else 
         ArgusLog (LOG_ERR, "ArgusSortQueue: ArgusMalloc(%d) %s\n", sizeof(struct ArgusRecord *), cnt, strerror(errno));
   }
#ifdef ARGUSDEBUG 
   ArgusDebug (5, "ArgusSortQueue(0x%x) returned\n", queue);
#endif
}


#ifdef ARGUS_CURSES
void RaUpdateWindow (struct ArgusParserStruct *, WINDOW *, struct ArgusQueueStruct *);

void
RaUpdateWindow (struct ArgusParserStruct *parser, WINDOW *window, struct ArgusQueueStruct *queue)
{
   struct ArgusRecordStruct *ns = NULL;
   int i, x, cnt, attr = A_NORMAL, z;
   short color = COLOR_WHITE;
   char buf[MAXSTRLEN], tbuf[MAXSTRLEN];

   if (RaWindowModified) {
      parser->RaLabel = NULL;
      if (RaWindowStatus) {
         if (queue->count < RaWindowStartLine) {
            RaWindowStartLine = queue->count - RaDisplayLines;
            RaWindowStartLine = (RaWindowStartLine > 0) ? RaWindowStartLine : 0;
         }

         cnt = ((RaDisplayLines > 0) ? RaDisplayLines : RaWindowLines);
         cnt = (cnt > (queue->count - RaWindowStartLine)) ? (queue->count - RaWindowStartLine) : cnt;
         attr = A_NORMAL;
         color = COLOR_WHITE;

         if (queue->array != NULL) {
            if ((parser->RaLabel == NULL) && (parser->ns != NULL)) {
               parser->RaLabel = ArgusGenerateLabel(parser, parser->ns);
               if (ArgusPrintRank)
                  snprintf (tbuf, COLS - 3, "Rank  %s", parser->RaLabel);
               else
                  snprintf (tbuf, COLS - 3, " %s", parser->RaLabel);
               mvwaddnstr (window, 0, 1, tbuf, COLS - 3);
               wclrtoeol(window);
            }

            for (x = 0, z = 0, i = RaWindowStartLine; x < cnt; x++, i++) {
               if ((ns = (struct ArgusRecordStruct *) queue->array[i]) != NULL) {
                  z++;
                  *(unsigned int *)buf = 0;
                  ArgusPrintRecord(parser, buf, ns, MAXSTRLEN);
                  if (ArgusPrintRank)
                     snprintf (tbuf, COLS - 3, "%4d %s", i + 1, buf);
                  else
                     snprintf (tbuf, COLS - 3, "%s", buf);

                  mvwaddnstr (window, x + 1, 2, tbuf, COLS - 3);
                  wclrtoeol(window);
               } else
                  break;
            }

            if (i < z)
               mvwaddnstr (window, x + 1, 0, " -more-\n", COLS - 3);

         } else
            mvwaddstr (window, 2, 0, " ");

         wclrtobot(window);
      }
      wnoutrefresh(window);
   }
   RaWindowModified = 0;
}
#endif

int ArgusSourceConnected = 0;

void
RaRefreshDisplay(struct ArgusParserStruct *parser)
{
   struct timeval startbuf, *start = &startbuf;
   struct timeval lastbuf, *last = &lastbuf;
   struct timeval tvp, dtime;
   char stimebuf[128], tbuf[MAXSTRLEN];
   char strbuf[128];  
   struct tm *tm, tmbuf;

   tvp = parser->ArgusRealTime;

   if (RaTopUpdateTime.tv_sec == 0)
      RaTopUpdateTime = tvp;
   
   if (RaWindowImmediate ||
      ((RaTopUpdateTime.tv_sec < tvp.tv_sec) ||
      ((RaTopUpdateTime.tv_sec == tvp.tv_sec) &&
       (RaTopUpdateTime.tv_usec <= tvp.tv_usec)))) {

#if defined(ARGUS_CURSES)
      RaUpdateWindow(parser, RaAvailableWindow, RaTopProcess->queue);
#endif
      RaTopUpdateTime = tvp;

      RaTopUpdateTime.tv_sec  += RaTopUpdateInterval.tv_sec;
      RaTopUpdateTime.tv_usec += RaTopUpdateInterval.tv_usec;

      if (RaTopUpdateTime.tv_usec >= 1000000) {
         RaTopUpdateTime.tv_sec  += 1;
         RaTopUpdateTime.tv_usec -= 1000000;
      }

#if defined(ARGUS_CURSES)
      RaWindowImmediate = FALSE;

      if (ArgusCurrentTime.tv_sec > 0) {
         tm = localtime_r((time_t *) &ArgusCurrentTime.tv_sec, &tmbuf);
         strftime ((char *) stimebuf, 32, "%Y/%m/%d.%T", tm);

         if ((!parser->Sflag) && (!RaTable)) {
            if (RaTopUpdateInterval.tv_usec > 0) {
               if ((RaTopUpdateInterval.tv_usec * RaUpdateRate) < 5) {
                  sprintf ((char *)&stimebuf[strlen(stimebuf)], ".%06d ", (int)(ArgusCurrentTime.tv_usec/100));
               } else
               if ((RaTopUpdateInterval.tv_usec * RaUpdateRate) < 50) {
                  sprintf ((char *)&stimebuf[strlen(stimebuf)], ".%05d ", (int)(ArgusCurrentTime.tv_usec/100));
               } else
               if ((RaTopUpdateInterval.tv_usec * RaUpdateRate) < 500) {
                  sprintf ((char *)&stimebuf[strlen(stimebuf)], ".%04d ", (int)(ArgusCurrentTime.tv_usec/100));
               } else
               if ((RaTopUpdateInterval.tv_usec * RaUpdateRate) < 5000) {
                  sprintf ((char *)&stimebuf[strlen(stimebuf)], ".%03d ", (int)(ArgusCurrentTime.tv_usec/1000));
               } else
               if ((RaTopUpdateInterval.tv_usec * RaUpdateRate) < 50000) {
                  sprintf ((char *)&stimebuf[strlen(stimebuf)], ".%02d ", (int)(ArgusCurrentTime.tv_usec/10000));
               } else
               if ((RaTopUpdateInterval.tv_usec * RaUpdateRate) < 500000) {
                  sprintf ((char *)&stimebuf[strlen(stimebuf)], ".%01d ", (int)(ArgusCurrentTime.tv_usec/100000));
               }
            }
         } else
            sprintf ((char *)&stimebuf[strlen(stimebuf)], " ");

         strftime(&stimebuf[strlen(stimebuf)], 32, "%z", tm);

      } else 
         sprintf (stimebuf, " ");

      mvwprintw (RaWindow, LINES - 1, 0, "\n");
      mvwaddnstr (RaHeaderWindow, 0, 0, ArgusGenerateProgramArgs(ArgusParser), COLS - 5);
      wclrtoeol(RaHeaderWindow);
      mvwaddnstr (RaHeaderWindow, 0, COLS - strlen(stimebuf) , stimebuf, strlen(stimebuf));

      if (ArgusPrintTotals) {
         if (parser->ns != NULL) {
            char buf[MAXSTRLEN];

            bzero (buf, 16);
            ArgusPrintRecord(parser, buf, parser->ns, MAXSTRLEN);
            if (ArgusPrintRank)
               snprintf (tbuf, COLS - 3, "%4d %s", RaTopProcess->queue->count, buf);
            else
               snprintf (tbuf, COLS - 3, "%s", buf);
         } else
            sprintf (tbuf, " ");

         mvwaddnstr (RaHeaderWindow, 3, 2, tbuf, COLS - 3);
         wclrtoeol(RaHeaderWindow);
      }

      if (parser->ArgusCurrentInput) {
         float secs, rate;
         char srcstr[128], *ptr;

         if (RaProbeUptime.tv_sec == 0) {
            start->tv_sec  = parser->ArgusCurrentInput->ArgusInitCon.argus_mar.startime.tv_sec;
            start->tv_usec = parser->ArgusCurrentInput->ArgusInitCon.argus_mar.startime.tv_usec;

            last->tv_sec   = parser->ArgusCurrentInput->ArgusInitCon.argus_mar.now.tv_sec;
            last->tv_usec  = parser->ArgusCurrentInput->ArgusInitCon.argus_mar.now.tv_usec;

            tvp.tv_sec  = last->tv_sec  - start->tv_sec;
            tvp.tv_usec = last->tv_usec - start->tv_usec;

            if (tvp.tv_usec < 0) {
               tvp.tv_sec--;
               tvp.tv_usec += 1000000;
            }
            if (tvp.tv_usec >= 1000000) {
               tvp.tv_sec++;
               tvp.tv_usec -= 1000000;
            }
            RaProbeUptime = tvp;
         }

         dtime.tv_sec   = RaTopStopTime.tv_sec  - RaTopStartTime.tv_sec;
         dtime.tv_usec  = RaTopStopTime.tv_usec - RaTopStartTime.tv_usec;

         if (dtime.tv_usec < 0) {
            dtime.tv_sec--;
            dtime.tv_usec += 1000000;
         }

         rate = (parser->ArgusTotalRecords * 1.0); 

         ptr = parser->ArgusCurrentInput->hostname ?
               parser->ArgusCurrentInput->hostname :
               parser->ArgusCurrentInput->filename ;

         if (!(ArgusSourceConnected)) {
            ArgusSourceConnected++;
            sprintf (parser->RaDebugString, "Source Connected %s", ptr);
         }

         if (ptr != NULL) {
            sprintf (srcstr, "%s", ptr);
            srcstr[64] = '\0';
         } else {
            srcstr[0] = '\0';
         }
         secs = (dtime.tv_sec * 1.0) + ((dtime.tv_usec * 1.0)/1000000.0);
         sprintf (tbuf, "Source %s  Version %d.%d  Queue %8d TotalRecords %8lld  Rate %11.4f rps",
                             srcstr, parser->ArgusCurrentInput->ArgusInitCon.argus_mar.major_version,
                             parser->ArgusCurrentInput->ArgusInitCon.argus_mar.minor_version,
                             RaTopProcess->queue->count,
                             parser->ArgusTotalRecords, rate/secs);

         mvwaddnstr (RaHeaderWindow, 2, 0, tbuf, COLS - 1);
         wclrtoeol(RaHeaderWindow);

      } else {
         if (ArgusSourceConnected) {
            ArgusSourceConnected = 0;
            sprintf (parser->RaDebugString, "No Source Connected");
         }
      }

      sprintf (strbuf, "%s", parser->RaDebugString);
      mvwaddnstr (RaWindow, LINES - 1, 1, strbuf, COLS - 3);

      wclrtoeol(RaWindow);
      wnoutrefresh(RaWindow);
#endif
   }

#if defined(ARGUS_CURSES)
   if (RaWindowStatus) {
      wclrtoeol(RaAvailableWindow);
      wnoutrefresh(RaAvailableWindow);
   }

   wnoutrefresh(RaHeaderWindow);

   if (RaCursorWindow == NULL)
      RaCursorWindow = RaHeaderWindow;

   wmove(RaCursorWindow, RaCursorY, RaCursorX - RaCursorOffset);
   wnoutrefresh(RaCursorWindow);
   doupdate();
#endif
}


int ArgusProcessQueue (struct ArgusQueueStruct *);
int ArgusProcessBins (struct ArgusRecordStruct *, struct RaBinProcessStruct *);
struct RaBinProcessStruct *ArgusNewRateBins (struct ArgusParserStruct *, struct ArgusRecordStruct *);


struct RaBinProcessStruct *
ArgusNewRateBins (struct ArgusParserStruct *parser, struct ArgusRecordStruct *ns)
{
   struct RaBinProcessStruct *retn = NULL;

   if ((retn = (struct RaBinProcessStruct *)ArgusCalloc(1, sizeof(*retn))) == NULL)
      ArgusLog (LOG_ERR, "ArgusNewRateBins: ArgusCalloc error %s", strerror(errno));
                        
   bcopy ((char *)RaBinProcess, (char *)retn, sizeof (*retn));

   retn->nadp.RaStartTmStruct = RaBinProcess->nadp.RaStartTmStruct;
   retn->nadp.RaEndTmStruct   = RaBinProcess->nadp.RaEndTmStruct;

   retn->startpt.tv_sec = mktime(&RaBinProcess->nadp.RaStartTmStruct);
   retn->endpt.tv_sec   = mktime(&RaBinProcess->nadp.RaEndTmStruct);

#if defined(ARGUSDEBUG)
   ArgusDebug (6, "ArgusNewRateBins (0x%x, 0x%x) returning %d", parser, ns, retn); 
#endif

   return(retn);
}


void ArgusShiftArray (struct ArgusParserStruct *, struct RaBinProcessStruct *);

int
ArgusProcessBins (struct ArgusRecordStruct *ns, struct RaBinProcessStruct *rbps)
{
   int retn = 0;
/*
   int i = 0, x = 0;
   struct ArgusRecordStruct *tns;
   struct RaBinStruct *bin;

   if (rbps->index > 0) {
      for (i = 0; i < rbps->index; i++) {
         if ((bin = rbps->array[i]) != NULL) {
            RaDeleteBin(ArgusParser, bin);
            rbps->array[i] = NULL;
         }
      }
   }
*/
   {
      int cnt   = (rbps->arraylen - rbps->index);
      int dtime = cnt * rbps->size;
      int rtime = (((ArgusParser->ArgusGlobalTime.tv_sec/rbps->size)) * rbps->size);

      if ((rbps->startpt.tv_sec + dtime) < rtime) {

         ArgusShiftArray(ArgusParser, rbps);

         RaWindowModified = RA_MODIFIED;
         RaWindowImmediate = TRUE;
         rbps->status |= RA_DIRTYBINS;
         retn = 1;
      }
   }

#if defined(ARGUSDEBUG)
   ArgusDebug (6, "ArgusProcessBins (0x%x, 0x%x) returning %d", ns, rbps, retn); 
#endif

   return (retn);
}


int
ArgusProcessQueue (struct ArgusQueueStruct *queue)
{
   struct ArgusHashTableHdr *htblhdr = NULL;
   struct timeval tbuf, *tvp = &tbuf;
   int retn = 0, x, z;

   if (ArgusParser->timeout) {
      struct ArgusRecordStruct *ns = (struct ArgusRecordStruct *)queue->start;
      struct ArgusRecordStruct *nxt = NULL;

      if (ns != NULL) {
         int count = queue->count;
         for (x = 0, z = count; x < z; x++) {
            *tvp = ns->qhdr.lasttime;
            tvp->tv_sec += ArgusParser->timeout;

            nxt = (struct ArgusRecordStruct *) ns->qhdr.nxt;

            if ((tvp->tv_sec  < ArgusParser->ArgusRealTime.tv_sec) ||
               ((tvp->tv_sec == ArgusParser->ArgusRealTime.tv_sec) &&
                (tvp->tv_usec < ArgusParser->ArgusRealTime.tv_usec))) {

               ArgusRemoveFromQueue(queue, &ns->qhdr);
               retn++;

               if (!(ns->status & ARGUS_NSR_STICKY)) {
                  if ((htblhdr = ns->htblhdr) != NULL)
                      ArgusRemoveHashEntry(htblhdr);

                  ArgusDeleteRecordStruct (ArgusParser, ns);

               } else
                  ArgusZeroRecord (ns);

            } else {
               struct RaBinProcessStruct *rbps;
               int i, z;

               if ((rbps = ns->bins) != NULL) {
                  ArgusProcessBins (ns, rbps);
                  if (rbps->status & RA_DIRTYBINS) {
                     ArgusZeroRecord (ns);
                     for (i = rbps->index; i < rbps->arraylen; i++) {
                        struct RaBinStruct *bin;
                        if (((bin = rbps->array[i]) != NULL) && (bin->queue != NULL)) {
                           struct ArgusRecordStruct *tns  = (struct ArgusRecordStruct *)bin->queue->start;
                           for (z = 0; z < bin->queue->count; z++) {
                              ArgusMergeRecords (ArgusAggregator, ns, tns);
                              tns = (struct ArgusRecordStruct *)tns->qhdr.nxt;
                           }
                        }
                     }

                     ns->status |= ARGUS_RECORD_MODIFIED;
                     rbps->status &= ~RA_DIRTYBINS;
                     retn++;
                  }
               }
            }

            if ((ns = nxt) == NULL)
               break;
         }
      }
   }

#if defined(ARGUSDEBUG)
   ArgusDebug (5, "ArgusProcessQueue (0x%x) returning %d", queue, retn); 
#endif

   return (retn);
}


extern void ArgusCloseInput(struct ArgusParserStruct *parser, struct ArgusInput *);
struct timeval RaProcessQueueTimer = {0, 250000};
void RaResizeScreen(void);
void ArgusUpdateScreen(void);

void
ArgusClientTimeout ()
{
   struct ArgusQueueStruct *queue = RaTopProcess->queue;
   struct timeval tvbuf, *tvp = &tvbuf;
   int i;

   gettimeofday(&ArgusParser->ArgusRealTime, 0);
   ArgusAdjustGlobalTime (ArgusParser, &ArgusParser->ArgusRealTime);
   *tvp = ArgusParser->ArgusGlobalTime;

   if (ArgusParser->RaClientUpdate.tv_sec != 0) {
      if (((ArgusParser->RaClientUpdate.tv_sec < tvp->tv_sec) ||
          ((ArgusParser->RaClientUpdate.tv_sec == tvp->tv_sec) &&
           (ArgusParser->RaClientUpdate.tv_usec < tvp->tv_usec)))) {

         if (ArgusProcessQueue(queue)) {
            RaWindowModified = RA_MODIFIED;
            RaWindowImmediate = TRUE;
         }

         ArgusParser->RaClientUpdate.tv_sec  += RaProcessQueueTimer.tv_sec;
         ArgusParser->RaClientUpdate.tv_usec += RaProcessQueueTimer.tv_usec;

         while (ArgusParser->RaClientUpdate.tv_usec > 1000000) {
            ArgusParser->RaClientUpdate.tv_sec++;
            ArgusParser->RaClientUpdate.tv_usec -= 1000000;
         }
      }

   if (RaWindowModified) {
      RaTopSortQueue(ArgusSorter, queue);
      if (ArgusParser->ns) {
         ArgusDeleteRecordStruct (ArgusParser, ArgusParser->ns);
         ArgusParser->ns = NULL;
      }
      for (i = 0; i < queue->count; i++) {
         struct ArgusRecordStruct *ns;

         if ((ns = (struct ArgusRecordStruct *)queue->array[i]) == NULL)
            break;

         if (ArgusParser->ns) 
            ArgusMergeRecords (ArgusAggregator, ArgusParser->ns, ns);
         else
            ArgusParser->ns = ArgusCopyRecordStruct (ns);
      }
   }

   ArgusUpdateScreen();

   } else
      ArgusParser->RaClientUpdate.tv_sec = ArgusParser->ArgusGlobalTime.tv_sec;

#if defined(ARGUSDEBUG)
   ArgusDebug (12, "ArgusClientTimeout () returning\n"); 
#endif
}

void
ArgusUpdateScreen(void)
{
   struct ArgusQueueStruct *queue = RaTopProcess->queue;

#if defined(ARGUS_CURSES)
   char RaOutputBuffer[MAXSTRLEN];
   int c, x, y;
   fd_set in;

   FD_ZERO(&in); FD_SET(0, &in);

   if (RaScreenLines == 0)
      RaScreenLines = LINES;
   if (RaScreenColumns == 0)
      RaScreenColumns = COLS;

   if ((RaScreenResize == TRUE) || ((RaScreenLines != LINES) || (RaScreenColumns != COLS)))
      RaResizeScreen();

   if (!(ArgusParser->RaTasksToDo)) {
      gettimeofday(&ArgusLastTime, 0);
      ArgusLastTime.tv_usec = 0;

   } else {
      if (ArgusCurrentTime.tv_sec != 0) {
         long long tUsec = 0;
         gettimeofday(&ArgusParser->ArgusRealTime, 0);

         if (ArgusLastRealTime.tv_sec > 0) {
            struct timeval dTime;

            dTime = *RaDiffTime(&ArgusParser->ArgusRealTime, &ArgusLastRealTime);
            tUsec = ((dTime.tv_sec * 1000000) + dTime.tv_usec) * RaUpdateRate;
            dTime.tv_sec  = tUsec / 1000000;
            dTime.tv_usec = tUsec % 1000000;

            ArgusCurrentTime.tv_sec  = ArgusLastTime.tv_sec  + dTime.tv_sec;
            ArgusCurrentTime.tv_usec = ArgusLastTime.tv_usec + dTime.tv_usec;

            if (ArgusCurrentTime.tv_usec > 1000000) {
               ArgusCurrentTime.tv_sec++;
               ArgusCurrentTime.tv_usec -= 1000000;
            }
         }
      }
   }

   switch (RaInputStatus) {
      case RANEWCOMMAND: {
         if ((c = wgetch(RaWindow)) != ERR) {
            RaWindowModified  = RA_MODIFIED;
            RaWindowImmediate = TRUE;
            mvwaddnstr (RaHeaderWindow, 1, 0, "  ", COLS - 2);
            wclrtoeol(RaHeaderWindow);
                
            if (RaCommandIndex != 0) { 
               RaCommandIndex = 0; 
               bzero (RaCommandInputStr, MAXSTRLEN);
               bzero (RaCommandError, MAXSTRLEN);
            }

            switch (c) {
               case 'k':
               case 0x19:
               case KEY_UP:
                  if (RaWindowStartLine > 0) {
                     RaWindowStartLine--;
                     wscrl(RaAvailableWindow, -1);
                     RaWindowModified++;
                  }
                  break;

               case 'j':
               case 0x05:
               case 0x0E:
               case KEY_DOWN:
                  if ((RaTopProcess->queue->count - RaWindowStartLine) > RaDisplayLines) {
                     RaWindowStartLine++;
                     wscrl(RaAvailableWindow, 1); 
                     RaWindowModified++;
                  } 
                  break;

               case 'g':
               case KEY_HOME:
                  if (RaWindowStartLine != 0) {
                     RaWindowStartLine = 0;
                     RaWindowModified++;
                  }
                  break;

               case 'G':
               case KEY_END:
                  if (RaWindowStartLine != (RaTopProcess->queue->count - RaDisplayLines)) {
                     RaWindowStartLine = RaTopProcess->queue->count - RaDisplayLines;
                     if (RaWindowStartLine < 0)
                        RaWindowStartLine = 0;
                     RaWindowModified++;
                  }
                  break;

               case 0x06:
               case 0x04:
               case ' ':
               case KEY_NPAGE:
                  if ((RaTopProcess->queue->count - RaWindowStartLine) > RaDisplayLines) {
                     RaWindowStartLine += RaDisplayLines;
                     wscrl(RaWindow, RaDisplayLines);
                     RaWindowModified++;
                  }
                  break;

               case 0x02:
               case 0x15:
               case KEY_PPAGE:
                  if (RaWindowStartLine > 0) { 
                     wscrl(RaWindow, (RaDisplayLines > RaWindowStartLine) ? -RaWindowStartLine : -RaDisplayLines);
                     RaWindowStartLine -= RaDisplayLines;
                     if (RaWindowStartLine < 0)
                        RaWindowStartLine = 0;
                     RaWindowModified++;
                  } 
                  break;

               case 0x0C:
                  werase(RaAvailableWindow);
                  RaWindowModified  = RA_MODIFIED;
                  RaWindowImmediate = TRUE;
                  RaRefreshDisplay(ArgusParser);
                  break;

               case 'a': {
                  RaInputStatus = RAGETTINGa;
                  RaInputString = RAGETTINGaSTR;
                  break;
               }

               case 'c': {
                  break;
               }

               case 'd': {
                  RaInputStatus = RAGETTINGd;
                  RaInputString = RAGETTINGdSTR;

                  if (ArgusParser->ArgusRemoteHostList) {
                     struct ArgusInput *input = ArgusParser->ArgusRemoteHostList;
                     while (input) {
                        sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], " %s:%d", input->hostname, input->portnum);
                        RaCommandIndex = strlen(RaCommandInputStr); 
                        input = input->nxt;
                     }
                  }

                  break;
               }
 
               case 'D': {
                  RaInputStatus = RAGETTINGD;
                  RaInputString = RAGETTINGDSTR;
                  sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], "%d", ArgusParser->debugflag);
                  RaCommandIndex = strlen(RaCommandInputStr); 
                  break;
               }

               case 'f': 
                  RaInputStatus = RAGETTINGf;
                  RaInputString = RAGETTINGfSTR;
                  RaFilterIndex = 3;
                  if (ArgusParser->ArgusRemoteFilter) {
                     sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], " remote %s", ArgusParser->ArgusRemoteFilter);
                     RaCommandIndex = strlen(RaCommandInputStr); 
                     RaFilterIndex = ARGUS_REMOTE_FILTER;
                  } else
                  if (ArgusParser->ArgusLocalFilter) {
                     sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], " local %s", ArgusParser->ArgusLocalFilter);
                     RaCommandIndex = strlen(RaCommandInputStr); 
                     RaFilterIndex = ARGUS_LOCAL_FILTER;
                  } else
                  if (ArgusParser->ArgusDisplayFilter) {
                     sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], " display %s", ArgusParser->ArgusDisplayFilter);
                     RaCommandIndex = strlen(RaCommandInputStr); 
                     RaFilterIndex = ARGUS_DISPLAY_FILTER;
                  }
                  break;

               case 'm': {
                  struct ArgusMaskStruct *ArgusMaskDefs = ArgusIpV4MaskDefs; 
                  int i;

                  RaInputStatus = RAGETTINGm;
                  RaInputString = RAGETTINGmSTR;

                  for (i = 0; i < ARGUS_MAX_MASK_LIST; i++) {
                     if (ArgusAggregator->mask & (0x01 << i)) {
                        sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], " %s", ArgusMaskDefs[i].name);

                        switch (i) {
                           case ARGUS_MASK_SADDR:
                              if (ArgusAggregator->saddrlen > 0)
                                 sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], "/%d", ArgusAggregator->saddrlen);
                              break;
                           case ARGUS_MASK_DADDR:
                              if (ArgusAggregator->daddrlen > 0)
                                 sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], "/%d", ArgusAggregator->daddrlen);
                              break;
                        }
                     }
                  }

                  RaCommandIndex = strlen(RaCommandInputStr);
                  break;
               }

               case 'M': {
                  struct ArgusModeStruct *mode;
                  RaInputStatus = RAGETTINGM;
                  RaInputString = RAGETTINGMSTR;
         
                  if ((mode = ArgusParser->ArgusModeList) != NULL) {
                     while (mode) {
                        sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], " %s", mode->mode);
                        mode = mode->nxt;
                     }
                  }
                  RaCommandIndex = strlen(RaCommandInputStr);
                  break;
               }

               case 'N':
                  RaInputStatus = RAGETTINGN;
                  RaInputString = RAGETTINGNSTR;
                  break;

               case 'p': {
                  ArgusParser->Pctflag = (ArgusParser->Pctflag == 1) ? 0 : 1;
                  if (ArgusParser->Pctflag)
                     RaInputString = "Toggle percent on";
                  else
                     RaInputString = "Toggle percent off";
                  break;
               }

               case 'P': {
                  double rate  = RaUpdateRate;
                  double pause = ArgusParser->Pauseflag;

                  ArgusParser->Pauseflag = (pause > 0.0) ? 0.0 : rate;
                  RaUpdateRate = (rate > 0.0) ? 0.0 : pause;

                  if (ArgusParser->Pauseflag)
                     RaInputString = "Paused";
                  else
                     RaInputString = "";
                  break;
               }

               case 't':
                  RaInputStatus = RAGETTINGt;
                  RaInputString = RAGETTINGtSTR;
                  if (ArgusParser->timearg) {
                     sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], "%s", ArgusParser->timearg);
                     RaCommandIndex = strlen(RaCommandInputStr); 
                  }
                  break;

               case 'T':
                  RaInputStatus = RAGETTINGT;
                  RaInputString = RAGETTINGTSTR;
                  sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], "%d", ArgusParser->timeout);
                  RaCommandIndex = strlen(RaCommandInputStr); 
                  break;

               case 'R': {
                  struct ArgusInput *input = ArgusParser->ArgusInputFileList;
                  RaInputStatus = RAGETTINGR;
                  RaInputString = RAGETTINGRSTR;
                  while (input) {
                     sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], " %s", input->filename);
                     RaCommandIndex = strlen(RaCommandInputStr); 
                     input = input->nxt;
                  }
                  break;
               }

               case 'r': {
                  struct ArgusInput *input = ArgusParser->ArgusInputFileList;
                  RaInputStatus = RAGETTINGr;
                  RaInputString = RAGETTINGrSTR;
                  while (input) {
                     sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], " %s", input->filename);
                     RaCommandIndex = strlen(RaCommandInputStr); 
                     input = input->nxt;
                  }
                  break;
               }

               case 'S': {
                  struct ArgusInput *input = ArgusParser->ArgusRemoteHostList;
                  RaInputStatus = RAGETTINGS;
                  RaInputString = RAGETTINGSSTR;
                  while (input) {
                     sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], " %s:%d", input->hostname, input->portnum);
                     RaCommandIndex = strlen(RaCommandInputStr); 
                     input = input->nxt;
                  }
                  break;
               }

               case 's': {
                  int x, y;
                  RaInputStatus = RAGETTINGs;
                  RaInputString = RAGETTINGsSTR;
                  for (x = 0; x < ARGUS_MAX_SORT_ALG; x++) {
                     if (ArgusSorter->ArgusSortAlgorithms[x]) {
                        for (y = 0; y < ARGUS_MAX_SORT_ALG; y++) {
                           if (ArgusSorter->ArgusSortAlgorithms[x] == ArgusSortAlgorithmTable[y]) {
                              sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], "%s ", 
                                    ArgusSortKeyWords[y]);
                              break;
                           }
                        }
                     }
                  }
                  RaCommandIndex = strlen(RaCommandInputStr); 
                  break;
               }

               case 'u':
                  RaInputStatus = RAGETTINGu;
                  RaInputString = RAGETTINGuSTR;
                  sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], "%d.", (int) RaTopUpdateInterval.tv_sec);
                  sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], "%06d",(int) RaTopUpdateInterval.tv_usec);
                  RaCommandIndex = strlen(RaCommandInputStr); 
                  break;

               case 'U':
                  RaInputStatus = RAGETTINGU;
                  RaInputString = RAGETTINGUSTR;
                  sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], "%2.2f", RaUpdateRate);
                  RaCommandIndex = strlen(RaCommandInputStr); 
                  break;

               case 'w':
                  RaInputStatus = RAGETTINGw;
                  RaInputString = RAGETTINGwSTR;
                  break;

               case 'F': 
                  RaInputStatus = RAGETTINGF;
                  RaInputString = RAGETTINGFSTR;

                  if (ArgusPrintRank)
                     sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], "rank ");

                  for (x = 0; x < MAX_PRINT_ALG_TYPES; x++) {
                     if (ArgusParser->RaPrintAlgorithms[x] != NULL) {
                        for (y = 0; y < MAX_PRINT_ALG_TYPES; y++) {
                           if ((void *) ArgusParser->RaPrintAlgorithms[x] == (void *) RaPrintAlgorithmTable[y].print) {
                              sprintf (&RaCommandInputStr[strlen(RaCommandInputStr)], "%s:%d ",
                                 RaPrintAlgorithmTable[y].field, RaPrintAlgorithmTable[y].length);
                              break;
                           }
                        }
                     } else
                        break;
                  }
                  RaCommandIndex = strlen(RaCommandInputStr);
                  break;

               case 'Q':
                  RaInputStatus = RAGETTINGq;
                  RaInputString = RAGETTINGqSTR;
                  break;

               case 'q':
                  RaInputStatus = RAGETTINGq;
                  RaInputString = RAGETTINGqSTR;
                  RaParseComplete(SIGINT);
                  break;

               case 'H':
                  ArgusParser->Hflag = ArgusParser->Hflag ? 0 : 1;
                  break;

               case 'h':
                  RaInputStatus = RAGETTINGh;
                  RaInputString = RAGETTINGhSTR;
                  RaWindowStatus = 0;
                  RaOutputHelpScreen();
                  break;

               case 'n':
                  if (++ArgusParser->nflag > 3) {
                     ArgusParser->nflag = 0;
                  }
                  break;

               case 'v': 
                  if (ArgusParser->vflag) {
                     ArgusParser->vflag = 0;
                     ArgusReverseSortDir = 0;
                  } else {
                     ArgusParser->vflag = 1;
                     ArgusReverseSortDir++;
                  }
                  RaTopSortQueue(ArgusSorter, RaTopProcess->queue);
                  break;

               case '=':  {
                  struct ArgusRecordStruct *ns = NULL;

                  werase(RaWindow);
                  RaWindowModified  = RA_MODIFIED;
                  RaWindowImmediate = TRUE;

                  while ((ns = (struct ArgusRecordStruct *) ArgusPopQueue(RaTopProcess->queue)) != NULL) 
                     ArgusDeleteRecordStruct (ArgusParser, ns);

                  ArgusEmptyHashTable(RaTopProcess->htable);
                  ArgusDeleteRecordStruct (ArgusParser, ArgusParser->ns);
                  ArgusParser->RaClientUpdate.tv_sec = 0;
                  ArgusParser->ArgusTotalRecords = 0;
                  RaTopStartTime.tv_sec = 0;
                  RaTopStartTime.tv_usec = 0;
                  RaTopStopTime.tv_sec = 0;
                  RaTopStopTime.tv_usec = 0;
                  ArgusParser->ns = NULL;
                  break;
               }

               case 'z':  
                  if (++ArgusParser->zflag > 1) {
                     ArgusParser->zflag = 0;
                  }
                  break;

               case 'Z':  
                  switch (ArgusParser->Zflag) {
                     case '\0': ArgusParser->Zflag = 'b'; break;
                     case  'b': ArgusParser->Zflag = 's'; break;
                     case  's': ArgusParser->Zflag = 'd'; break;
                     case  'd': ArgusParser->Zflag = '\0'; break;
                  }
                  break;

               default:
                  break;
            }
         }
         break;
      }

      case RAGETTINGa:
      case RAGETTINGA:
      case RAGETTINGB:
      case RAGETTINGc:
      case RAGETTINGd:
      case RAGETTINGD:
      case RAGETTINGf:
      case RAGETTINGt:
      case RAGETTINGT:
      case RAGETTINGR:
      case RAGETTINGr:
      case RAGETTINGs:
      case RAGETTINGS:
      case RAGETTINGu:
      case RAGETTINGU:
      case RAGETTINGN:
      case RAGETTINGw:
      case RAGETTINGF:
      case RAGETTINGq:
      case RAGETTINGh:
      case RAGETTINGm:
      case RAGETTINGM: {
         struct timeval tvbuf, *tvp = &tvbuf;
         int ch;

         tvp->tv_sec = 0; tvp->tv_usec = 0;
         while (select(1, &in, 0, 0, tvp) > 0) {
            if ((ch = wgetch(RaWindow)) != ERR) {
               switch (RaInputStatus) {
                  case RAGETTINGh: 
                     RaWindowStatus = 1;
                     werase(RaAvailableWindow);
                     RaInputStatus = RANEWCOMMAND;
                     RaInputString = RANEWCOMMANDSTR;
                     RaCommandInputStr[0] = '\0';
                     RaWindowModified++;
                     break;

                  default: {
                     if ((ch == '\n') || (ch == '\r')) {
                        RaCursorOffset = 0;
                        RaCommandInputStr[RaCommandIndex] = '\0';
                        switch (RaInputStatus) {
                           case RAGETTINGN: {
                              char *ptr = NULL;
                              int value = strtol(RaCommandInputStr, (char **)&ptr, 10);

                              if (ptr != RaCommandInputStr) {
                                 RaDisplayLines = ((value < (LINES - (RaHeaderWinSize + 1)) - 1) ?
                                                    value : (LINES - (RaHeaderWinSize + 1)) - 1);
                                 RaWindowModified++;
                                 RaWindowImmediate = TRUE;
                              }
                        
                              break;
                           }

                           case RAGETTINGS: {
                              struct ArgusInput *input = NULL;
                              if (!(ArgusAddHostList (ArgusParser, RaCommandInputStr, (ArgusParser->Cflag ? ARGUS_CISCO_DATA_SOURCE : ARGUS_DATA_SOURCE)))) {
                                 ArgusLog (LOG_ALERT, "%s%s host not found", RaInputString, RaCommandInputStr);
                              } else {
                                 ArgusDeleteHostList(ArgusParser);
                                 ArgusAddHostList (ArgusParser, RaCommandInputStr, (ArgusParser->Cflag ? ARGUS_CISCO_DATA_SOURCE : ARGUS_DATA_SOURCE));
                                 ArgusParser->Sflag = 1;
                                 ArgusParser->RaParseDone = 0;
                              }

                              if ((input = ArgusParser->ArgusRemoteHostList) != NULL) {
                                 while (input) {
                                    if ((input->fd = ArgusGetServerSocket (input, 5)) >= 0) {
                                       if ((ArgusReadConnection (ArgusParser, input, ARGUS_SOCKET)) >= 0) {
                                          ArgusParser->ArgusRemoteFDs[ArgusParser->ArgusActiveFDs++] = input;
                                          ArgusParser->RaTasksToDo = 1;
                                       } else {
                                          sprintf (ArgusParser->RaDebugString, "ArgusReadConnection '%s': %s", input->hostname, strerror(errno));
                                       }

                                    } else {
                                       sprintf (ArgusParser->RaDebugString, "ArgusGetServerSocket '%s': %s", input->hostname, strerror(errno));
                                    }
                                    input = input->nxt;
                                 }
                              }
                              break;
                           }

                           case RAGETTINGa: {
                              if (!(strncasecmp(RaCommandInputStr, "Totals", 6))) {
                                 RaScreenResize = TRUE;
                                 ArgusPrintTotals++;
                                 RaHeaderWinSize++;
                                 RaWindowModified++;
                                 RaWindowImmediate = TRUE;
                              }
                           }
                           break;

                           case RAGETTINGd: {
                              int del = 0, n, i;
                              char strbuf[MAXSTRLEN];

                              if ((n = ArgusParser->ArgusActiveFDs) > 0) {
                                 for (i = 0; i < n; i++) {
                                    struct ArgusInput *input;

                                    if ((input = ArgusParser->ArgusRemoteFDs[i]) != NULL) {
                                       sprintf (strbuf, " %s:%d", input->hostname, input->portnum);

                                       if ((strstr (RaCommandInputStr, strbuf))) {
                                          del++;
                                          ArgusCloseInput(ArgusParser, input);
                                          ArgusParser->ArgusRemoteFDs[i] = NULL;
                                          if (ArgusParser->ArgusCurrentInput == input)
                                             ArgusParser->ArgusCurrentInput = NULL;
                                       }
                                    }
                                 }

                                 if (del > 0) {
                                    for (i = 0; i < n; i++) {
                                       while (n && (ArgusParser->ArgusRemoteFDs[i] == NULL)) {
                                          for (x = i + 1; x < n; x++) {
                                             ArgusParser->ArgusRemoteFDs[i] = ArgusParser->ArgusRemoteFDs[x];
                                             ArgusParser->ArgusRemoteFDs[x] = NULL;
                                          }
                                          n--;
                                       }
                                    }
                                 }

                                 if (ArgusParser->ArgusActiveFDs == del) {
                                    ArgusDeleteHostList(ArgusParser);
                                    ArgusParser->RaTasksToDo = 0;
                                    ArgusParser->Sflag = 0;

                                 }

                                 ArgusParser->ArgusActiveFDs -= del;
                              }
                           }
                           break;

                           case RAGETTINGD: {
                              char *ptr = NULL;
                              int value = strtol(RaCommandInputStr, (char **)&ptr, 10);

                              if (ptr != RaCommandInputStr)
                                 ArgusParser->debugflag = value;
                              break;
                           }

                           case RAGETTINGc: {
                              break;
                           }

                           case RAGETTINGf: {
                              struct nff_program lfilter;
                              char *ptr, *str = NULL;
                              int ind = ARGUS_REMOTE_FILTER;
                              int i, retn;

                              bzero ((char *) &lfilter, sizeof (lfilter));
                              ptr = RaCommandInputStr;
                              while (isspace((int)*ptr)) ptr++;

                              if ((str = strstr (ptr, "local")) != NULL) {
                                 ptr = strdup(&str[strlen("local ")]);
                                 ind = ARGUS_LOCAL_FILTER;
                              } else 
                              if ((str = strstr (ptr, "display")) != NULL) {
                                 ptr = strdup(&str[strlen("display ")]);
                                 ind = ARGUS_DISPLAY_FILTER;
                              } else 
                              if ((str = strstr (ptr, "remote")) != NULL) {
                                 ptr = strdup(&str[strlen("remote ")]);
                                 ind = ARGUS_REMOTE_FILTER;
                              } else 
                              if ((str = strstr (ptr, "none")) != NULL) {
                                 ind = RaFilterIndex;
                              }

                              if ((retn = ArgusFilterCompile (&lfilter, ptr, 1)) < 0)
                                 sprintf (ArgusParser->RaDebugString, "%s %s syntax error", RAGETTINGfSTR, RaCommandInputStr);

                              else {
                                 sprintf (ArgusParser->RaDebugString, "%s %s filter accepted", RAGETTINGfSTR, RaCommandInputStr);
                                 str = ptr;
                                 while (isspace((int)*str)) str++;
                                 
                                 switch (ind) {
                                    case ARGUS_LOCAL_FILTER:
                                       if (ArgusParser->ArgusFilterCode.bf_insns != NULL)
                                          free (ArgusParser->ArgusFilterCode.bf_insns);

                                       bcopy((char *)&lfilter, (char *)&ArgusParser->ArgusFilterCode, sizeof(lfilter));
                                       if (ArgusParser->ArgusLocalFilter !=  NULL) {
                                          free(ArgusParser->ArgusLocalFilter);
                                          ArgusParser->ArgusLocalFilter = NULL;
                                       }
                                       if (strlen(str) > 0)
                                          ArgusParser->ArgusLocalFilter = ptr;
                                       else
                                          free(ptr);
                                       break;

                                    case ARGUS_DISPLAY_FILTER:
                                       if (ArgusParser->ArgusDisplayCode.bf_insns != NULL)
                                          free (ArgusParser->ArgusDisplayCode.bf_insns);

                                       bcopy((char *)&lfilter, (char *)&ArgusParser->ArgusDisplayCode, sizeof(lfilter));
                                       bcopy((char *)&lfilter, (char *)&ArgusSorter->filter, sizeof(lfilter));

                                       if (ArgusParser->ArgusDisplayFilter !=  NULL) {
                                          free(ArgusParser->ArgusDisplayFilter);
                                          ArgusParser->ArgusDisplayFilter = NULL;
                                       }
                                       if (strlen(str) > 0)
                                          ArgusParser->ArgusDisplayFilter = ptr;
                                       else
                                          free(ptr);
                                       break;

                                    case ARGUS_REMOTE_FILTER:
                                       if (ArgusParser->ArgusFilterCode.bf_insns != NULL)
                                          free (ArgusParser->ArgusFilterCode.bf_insns);
                                       bcopy((char *)&lfilter, (char *)&ArgusParser->ArgusFilterCode, sizeof(lfilter));
                                       if (ArgusParser->ArgusRemoteFilter !=  NULL) {
                                          free(ArgusParser->ArgusRemoteFilter);
                                          ArgusParser->ArgusRemoteFilter = NULL;
                                       }
                                       if (strlen(str) > 0)
                                          ArgusParser->ArgusRemoteFilter = ptr;
                                       else
                                          free(ptr);
                                       break;
                                 }
                              }

                              RaTopSortQueue(ArgusSorter, RaTopProcess->queue);
                              if (ArgusParser->ns) {
                                 ArgusDeleteRecordStruct (ArgusParser, ArgusParser->ns);
                                 ArgusParser->ns = NULL;
                              }
                              for (i = 0; i < queue->count; i++) {
                                 if (queue->array[i] == NULL)
                                    break;
                                 if (ArgusParser->ns)
                                    ArgusMergeRecords (ArgusAggregator, ArgusParser->ns, (struct ArgusRecordStruct *)queue->array[i]);
                                 else
                                    ArgusParser->ns = ArgusCopyRecordStruct ((struct ArgusRecordStruct *)queue->array[i]);
                              }

                              RaWindowModified++;
                              RaWindowImmediate = TRUE;
                              break;
                           }
                                        
                           case RAGETTINGm: {
                              struct ArgusRecordStruct *ns = NULL;
                              char strbuf[MAXSTRLEN], *tok = NULL, *ptr;
                              struct ArgusModeStruct *mode = NULL, *modelist = NULL, *list; 
                              int i;                                  

                              ArgusParser->RaMonMode = 0;
                              strncpy(strbuf, RaCommandInputStr, MAXSTRLEN);

                              if ((mode = ArgusParser->ArgusMaskList) != NULL)
                                 ArgusDeleteMaskList(ArgusParser);

                              ArgusAggregator->mask = 0;

                              if ((ptr = strbuf) != NULL) {
                                 while ((tok = strtok (ptr, " \t")) != NULL) {
                                    if ((mode = (struct ArgusModeStruct *) ArgusCalloc (1, sizeof(struct ArgusModeStruct))) != NULL) {
                                       if ((list = modelist) != NULL) {
                                          while (list->nxt)
                                             list = list->nxt;
                                          list->nxt = mode;
                                       } else
                                          modelist = mode;
                                       mode->mode = strdup(tok);
                                    }
                                    ptr = NULL;
                                 }
                              } else {
                                 if ((modelist = ArgusParser->ArgusMaskList) == NULL)
                                    ArgusAggregator->mask  = ( ARGUS_MASK_SRCID_INDEX | ARGUS_MASK_PROTO_INDEX |
                                                              ARGUS_MASK_SADDR_INDEX | ARGUS_MASK_SPORT_INDEX |
                                                              ARGUS_MASK_DADDR_INDEX | ARGUS_MASK_DPORT_INDEX );
                              }

                              ArgusInitAggregatorStructs(ArgusAggregator);

                              if ((mode = modelist) != NULL) {
                                 while (mode) {
                                    char *ptr = NULL, **endptr = NULL;
                                    int value = 0;

                                    if ((ptr = strchr(mode->mode, '/')) != NULL) {
                                       *ptr++ = '\0';
                                       if ((value = strtol(ptr, endptr, 10)) == 0)
                                          if (*endptr == ptr)
                                             usage();
                                    }
                                    if (!(strncasecmp (mode->mode, "none", 4))) {
                                       ArgusAggregator->mask  = 0;
                                    } else
                                    if (!(strncasecmp (mode->mode, "mac", 3))) {
                                       ArgusParser->RaMonMode++;
                                       ArgusAggregator->mask |= (0x01 << ARGUS_MASK_SMAC);
                                       if (value > 0) {
                                          ArgusAggregator->saddrlen = value;
                                          ArgusAggregator->daddrlen = value;
                                       }
                                    } else
                                    if (!(strncasecmp (mode->mode, "addr", 4))) {
                                       ArgusParser->RaMonMode++;
                                       ArgusAggregator->mask |= (0x01 << ARGUS_MASK_SADDR);
                                       if (value > 0) {
                                          ArgusAggregator->saddrlen = value;
                                          ArgusAggregator->daddrlen = value;
                                       }
                                    } else
                                    if (!(strncasecmp (mode->mode, "matrix", 6))) {
                                       ArgusAggregator->ArgusMatrixMode++;
                                       ArgusAggregator->mask |= (0x01 << ARGUS_MASK_SADDR);
                                       ArgusAggregator->mask |= (0x01 << ARGUS_MASK_DADDR);
                                       if (value > 0) {
                                          ArgusAggregator->saddrlen = value;
                                          ArgusAggregator->daddrlen = value;
                                       }
                                    } else {
                                       struct ArgusMaskStruct *ArgusMaskDefs = ArgusIpV4MaskDefs;

                                       for (i = 0; i < ARGUS_MAX_MASK_LIST; i++) {
                                          if (!(strncasecmp (mode->mode, ArgusMaskDefs[i].name, ArgusMaskDefs[i].slen))) {
                                             ArgusAggregator->mask |= (0x01 << i);
                                             switch (i) {
                                                case ARGUS_MASK_SADDR:
                                                   if (value > 0)
                                                      ArgusAggregator->saddrlen = value;
                                                   break;
                                                case ARGUS_MASK_DADDR:
                                                   if (value > 0)
                                                      ArgusAggregator->daddrlen = value;
                                                   break;

                                                case ARGUS_MASK_SMPLS:
                                                case ARGUS_MASK_DMPLS: {
                                                   int x, RaNewIndex = 0;
                                                   char *ptr;

                                                   if ((ptr = strchr(mode->mode, '[')) != NULL) {
                                                      char *cptr = NULL;
                                                      int sind = -1, dind = -1;
                                                      *ptr++ = '\0';
                                                      while (*ptr != ']') {
                                                         if (isdigit((int)*ptr)) {
                                                            dind = strtol(ptr, (char **)&cptr, 10);
                                                            if (cptr == ptr)
                                                               usage ();
               
                                                            if (sind < 0)
                                                               sind = dind;

                                                            for (x = sind; x <= dind; x++)
                                                               RaNewIndex |= 0x01 << x;

                                                            ptr = cptr;
                                                            if (*ptr != ']')
                                                               ptr++;
                                                            if (*cptr != '-')
                                                               sind = -1;
                                                         } else
                                                            usage ();
                                                      }
                                                      ArgusIpV4MaskDefs[i].index = RaNewIndex;
                                                      ArgusIpV6MaskDefs[i].index = RaNewIndex;
                                                      ArgusEtherMaskDefs[i].index = RaNewIndex;
                                                   }
                                                   break;
                                                }
                                             }
                                             break;
                                          }
                                       }
                                    }
                                    mode = mode->nxt;
                                 }
                              }

                              while ((ns = (struct ArgusRecordStruct *) ArgusPopQueue(RaTopProcess->queue)) != NULL)
                                 ArgusDeleteRecordStruct (ArgusParser, ns);

                              ArgusEmptyHashTable(RaTopProcess->htable);
                              ArgusDeleteRecordStruct (ArgusParser, ArgusParser->ns);
                              ArgusParser->RaClientUpdate.tv_sec = 0;
                              ArgusParser->ns = NULL;

                              break;
                           }

                           case RAGETTINGM: {
                              struct ArgusModeStruct *mode = NULL;
                              char strbuf[MAXSTRLEN], *str = strbuf, *tok = NULL;
                              char *tzptr;

                              strncpy(strbuf, RaCommandInputStr, MAXSTRLEN);

                              if ((tzptr = strstr(strbuf, "TZ=")) != NULL) {
                                 if (ArgusParser->RaTimeZone != NULL)
                                    free (ArgusParser->RaTimeZone);
                                 ArgusParser->RaTimeZone = strdup(tzptr + 3);
                                 setenv("TZ", ArgusParser->RaTimeZone, 1);
                                 tzset();

                                 RaWindowModified  = RA_MODIFIED;
                                 RaWindowImmediate = TRUE;
                                 break;
                              }


                              if ((mode = ArgusParser->ArgusModeList) != NULL) {
                                 ArgusDeleteModeList(ArgusParser);
                                 ArgusParser->RaCumulativeMerge = 1;
                              }

                              if (strlen(strbuf) > 0) {
                                 while ((tok = strtok(str, " \t\n")) != NULL) {
                                    if (!(strncasecmp (tok, "none", 4)))
                                       ArgusDeleteModeList(ArgusParser);
                                    else if (!(strncasecmp (tok, "default", 7))) {
                                       ArgusDeleteModeList(ArgusParser);
                                    } else
                                       ArgusAddModeList (ArgusParser, tok);
                                    str = NULL;
                                 }
                              }

                              if ((mode = ArgusParser->ArgusModeList) != NULL) {
                                 struct ArgusAdjustStruct *nadp = NULL;
                                 int i, ind;

                                 while (mode) {
                                    for (i = 0, ind = -1; i < ARGUSSPLITMODENUM; i++) {
                                       if (!(strncasecmp (mode->mode, RaSplitModes[i], strlen(RaSplitModes[i])))) {
                                          ind = i;
                                          break;
                                       }
                                    }

                                    if (ind >= 0) {
                                       char *mptr = NULL;
                                       int size = -1;
                                       nadp = &RaBinProcess->nadp;

                                       nadp = &RaBinProcess->nadp;

                                       switch (ind) {
                                          case ARGUSSPLITRATE:  {   /* "%d:%d[yMwdhms]" */
                                             struct ArgusModeStruct *tmode = NULL; 
                                             nadp->mode = ind;
                                             if ((tmode = mode->nxt) != NULL) {
                                                mptr = tmode->mode;
                                                if (isdigit((int)*tmode->mode)) {
                                                   char *ptr = NULL;
                                                   nadp->len = strtol(tmode->mode, (char **)&ptr, 10);
                                                   if (*ptr++ != ':') 
                                                      usage();
                                                   tmode->mode = ptr;
                                                }
                                             }
                                          }

                                          case ARGUSSPLITTIME: /* "%d[yMwdhms] */
                                             nadp->mode = ind;
                                             if ((mode = mode->nxt) != NULL) {
                                                if (isdigit((int)*mode->mode)) {
                                                   char *ptr = NULL;
                                                   nadp->value = strtol(mode->mode, (char **)&ptr, 10);
                                                   if (ptr == mode->mode)
                                                      usage();
                                                   else {
                                                      switch (*ptr) {
                                                         case 'y':
                                                            nadp->qual = ARGUSSPLITYEAR;  
                                                            size = nadp->value * 31556926;
                                                            break;
                                                         case 'M':
                                                            nadp->qual = ARGUSSPLITMONTH; 
                                                            size = nadp->value * 2629744;
                                                            break;
                                                         case 'w':
                                                            nadp->qual = ARGUSSPLITWEEK;  
                                                            size = nadp->value * 604800;
                                                            break;
                                                         case 'd':
                                                            nadp->qual = ARGUSSPLITDAY;   
                                                            size = nadp->value * 86400;
                                                            break;
                                                         case 'h':
                                                            nadp->qual = ARGUSSPLITHOUR;  
                                                            size = nadp->value * 3600;
                                                            break;
                                                         case 'm':
                                                            nadp->qual = ARGUSSPLITMINUTE;
                                                            size = nadp->value * 60;
                                                            break;
                                                          default:
                                                            nadp->qual = ARGUSSPLITSECOND;
                                                            size = nadp->value;
                                                            break;
                                                      }
                                                   }
                                                }
                                                if (mptr != NULL)
                                                    mode->mode = mptr;
                                             }

                                             nadp->modify = 1;

                                             if (ind == ARGUSSPLITRATE) {
                                                /* need to set the flow idle timeout value to be equal to or
                                                   just a bit bigger than (nadp->len * size) */

                                                ArgusParser->timeout = (nadp->len * size);
                                             }

                                             ArgusSorter->ArgusSortAlgorithms[0] = ArgusSortStartTime;
                                             ArgusSorter->ArgusSortAlgorithms[1] = NULL;
                                             break;

                                          case ARGUSSPLITSIZE:
                                          case ARGUSSPLITCOUNT:
                                             nadp->mode = ind;
                                             nadp->count = 1;

                                             if ((mode = mode->nxt) != NULL) {
                                                if (isdigit((int)*mode->mode)) {
                                                   char *ptr = NULL;
                                                   nadp->value = strtol(mode->mode, (char **)&ptr, 10);
                                                   if (ptr == mode->mode)
                                                      usage();
                                                   else {
                                                      switch (*ptr) {
                                                         case 'B':   
                                                         case 'b':  nadp->value *= 1000000000; break;
                                                          
                                                         case 'M':   
                                                         case 'm':  nadp->value *= 1000000; break;
                                                          
                                                         case 'K':   
                                                         case 'k':  nadp->value *= 1000; break;
                                                      }
                                                   }
                                                }
                                             }
                                             ArgusSorter->ArgusSortAlgorithms[0] = NULL;
                                             break;

                                          case ARGUSSPLITNOMODIFY:
                                             nadp->modify = 0;
                                             break;

                                          case ARGUSSPLITSOFT:
                                             nadp->soft++;
                                             break;

                                          case ARGUSSPLITZERO:
                                             nadp->zero++;
                                             break;
                                       }

                                    } else {
                                       if (!(strncasecmp (mode->mode, "nomerge", 7))) {
                                          ArgusParser->RaCumulativeMerge = 0;
                                       } else
                                       if (!(strncasecmp (mode->mode, "merge", 5))) {
                                          ArgusParser->RaCumulativeMerge = 1;
                                       } else
                                       if (!(strncasecmp (mode->mode, "rtime", 5)) ||
                                          (!(strncasecmp (mode->mode, "realtime", 8)))) {
                                          char *ptr = NULL;
                                          RaTopRealTime++;
                                          if ((ptr = strchr(mode->mode, ':')) != NULL) {
                                             double value = 0.0;
                                             char *endptr = NULL;
                                             ptr++;
                                             value = strtod(ptr, &endptr);
                                             if (ptr != endptr) {
                                                RaUpdateRate = value;
                                             }
                                          }

                                       }
                                    }

                                    mode = mode->nxt;
                                 }
                              }

                              break;
                           }

                           case RAGETTINGR: {
                              char strbuf[MAXSTRLEN], *str = strbuf, *ptr = NULL;
                              strncpy(strbuf, RaCommandInputStr, MAXSTRLEN);

                              if (strlen(strbuf) > 0) {
                                 ArgusDeleteFileList(ArgusParser);
                                 while ((ptr = strtok(str, " ")) != NULL) {
                                    RaProcessRecursiveFiles (ptr);
                                    str = NULL;
                                 }
                              }
                              break;
                           }

                           case RAGETTINGr: {
                              char strbuf[MAXSTRLEN], *str = strbuf, *ptr = NULL;
                              glob_t globbuf;

                              bzero (strbuf, MAXSTRLEN);
                              strncpy(strbuf, RaCommandInputStr, MAXSTRLEN);

                              if (strlen(strbuf) > 0) {
                                 struct ArgusRecordStruct *ns = NULL;

                                 ArgusDeleteFileList(ArgusParser);
                                 while ((ptr = strtok(str, " ")) != NULL) {
                                    glob (ptr, 0, NULL, &globbuf);
                                    if (globbuf.gl_pathc > 0) {
                                       int i;
                                       for (i = 0; i < globbuf.gl_pathc; i++)
                                          ArgusAddFileList (ArgusParser, globbuf.gl_pathv[i], ARGUS_DATA_SOURCE, -1, -1);
                                    } else 
                                       sprintf (ArgusParser->RaDebugString, "%s no files found for %s", RAGETTINGrSTR, ptr);
                                    str = NULL;
                                 }
                                 ArgusParser->RaTasksToDo = 1;
                                 ArgusParser->Sflag = 0;
                                 while ((ns = (struct ArgusRecordStruct *) ArgusPopQueue(RaTopProcess->queue)) != NULL) 
                                    ArgusDeleteRecordStruct (ArgusParser, ns);
                                 ArgusEmptyHashTable(RaTopProcess->htable);
                                 ArgusDeleteRecordStruct (ArgusParser, ArgusParser->ns);
                                 ArgusParser->RaClientUpdate.tv_sec = 0;
                                 ArgusParser->status &= ~ARGUS_FILE_LIST_PROCESSED;
                                 ArgusParser->ns = NULL;
                                 ArgusLastTime.tv_sec  = 0;
                                 ArgusLastTime.tv_usec = 0;
                              }
                              break;
                           }

                           case RAGETTINGs: {
                              char strbuf[MAXSTRLEN], *ptr = strbuf, *tok;
                              int (*srtalg[ARGUS_MAX_SORT_ALG])(struct ArgusRecordStruct *, struct ArgusRecordStruct *);
                              int x, ind = 0;
                              strncpy (strbuf, RaCommandInputStr, MAXSTRLEN);
                              bzero(srtalg, sizeof(srtalg));
                              while ((tok = strtok(ptr, " ")) != NULL) {
                                 for (x = 0; x < ARGUS_MAX_SORT_ALG; x++) {
                                    if (!strncmp (ArgusSortKeyWords[x], tok, strlen(ArgusSortKeyWords[x]))) {
                                       srtalg[ind++] = ArgusSortAlgorithmTable[x];
                                       break;
                                    }
                                 }
                                 if (x == ARGUS_MAX_SORT_ALG) {
                                    bzero(srtalg, sizeof(srtalg));
                                    ArgusLog (LOG_ALERT, "sort keyword %s not valid", tok);
                                    break;
                                 }
                                 ptr = NULL;
                              }

                              if (srtalg[0] != NULL) {
                                 for (x = 0; x < ARGUS_MAX_SORT_ALG; x++)
                                    ArgusSorter->ArgusSortAlgorithms[x] = srtalg[x];
                                 
                              }
                              RaTopSortQueue(ArgusSorter, RaTopProcess->queue);
                              RaWindowModified  = RA_MODIFIED;
                              RaWindowImmediate = TRUE;
                              break;
                           }

                           case RAGETTINGT: {
                              char *ptr = NULL;
                              int value = strtol(RaCommandInputStr, (char **)&ptr, 10);

                              if (ptr != RaCommandInputStr)
                                 ArgusParser->timeout = value;
                              break;
                           }

                           case RAGETTINGt: {
                              if (ArgusParser->timearg) {
                                 free (ArgusParser->timearg);
                                 ArgusParser->timearg = NULL;
                              }

                              if (RaCommandInputStr)
                                 if (strlen(RaCommandInputStr))
                                    ArgusParser->timearg = strdup(RaCommandInputStr);

                              ArgusCheckTimeFormat (ArgusParser->RaTmStruct, ArgusParser->timearg);
                              break;
                           }

                           case RAGETTINGu: {
                              double value = 0.0, ivalue, fvalue;
                              char *endptr = NULL;
 
                              value = strtod(RaCommandInputStr, &endptr);
 
                              if (RaCommandInputStr != endptr) {
                                 fvalue = modf(value, &ivalue);
 
                                 RaTopUpdateInterval.tv_sec  = (int) ivalue;
                                 RaTopUpdateInterval.tv_usec = (int) (fvalue * 1000000.0);
 
                                 sprintf (ArgusParser->RaDebugString, "%s %s interval accepted", RAGETTINGuSTR, RaCommandInputStr);
                                 RaTopUpdateTime = ArgusParser->ArgusRealTime;
 
                              } else
                                 sprintf (ArgusParser->RaDebugString, "%s %s syntax error", RAGETTINGuSTR, RaCommandInputStr);

                              break;
                           }


                           case RAGETTINGU: {
                              double value = 0.0;
                              char *endptr = NULL;
 
                              value = strtod(RaCommandInputStr, &endptr);
 
                              if (RaCommandInputStr != endptr) {
                                 RaUpdateRate = value;
                                 sprintf (ArgusParser->RaDebugString, "%s %s accepted", RAGETTINGUSTR, RaCommandInputStr);
 
                              } else
                                 sprintf (ArgusParser->RaDebugString, "%s %s syntax error", RAGETTINGUSTR, RaCommandInputStr);

                              break;
                           }

                           
                           case RAGETTINGw: {
                              struct ArgusListStruct *wlist = ArgusParser->ArgusWfileList;
                              struct ArgusWfileStruct *wfile = NULL;
                              struct ArgusRecord *argusrec = NULL;
                              struct ArgusRecordStruct *ns;
                              static char sbuf[0x10000];
                              int i;

                              if (RaTopProcess->queue->count > 0) {
                                 ArgusParser->ArgusWfileList = NULL;
                                 setArgusWfile (ArgusParser, RaCommandInputStr, NULL);
                                 wfile = (struct ArgusWfileStruct *) ArgusParser->ArgusWfileList->start;
                                 RaTopSortQueue(ArgusSorter, RaTopProcess->queue);
                                 for (i = 0; i < RaTopProcess->queue->count; i++) {
                                    if ((ns = (struct ArgusRecordStruct *) queue->array[i]) == NULL)
                                       break;

                                    if ((argusrec = ArgusGenerateRecord (ns, 0L, sbuf)) != NULL) {
#ifdef _LITTLE_ENDIAN
                                       ArgusHtoN(argusrec);
#endif
                                       ArgusWriteNewLogfile (ArgusParser, wfile, argusrec);

                                    }
                                 }
                              
                                 fflush(wfile->fd);
                                 fclose(wfile->fd);
                                 clearArgusWfile(ArgusParser);
                                 ArgusParser->ArgusWfileList = wlist;
                              }

                              break;   
                           }

                           case RAGETTINGF: {
                              char strbuf[MAXSTRLEN], *ptr = strbuf, *tok;
                              int x;
                              strncpy (strbuf, RaCommandInputStr, MAXSTRLEN);
                              bzero ((char *)ArgusParser->RaSOptionStrings, sizeof(ArgusParser->RaSOptionStrings));
                              ArgusParser->RaSOptionIndex = 0;
                              ArgusPrintRank = 0;
                              while ((tok = strtok(ptr, " ")) != NULL) {
                                 if ((strstr (tok, "rank"))) {
                                    if (*tok == '-')
                                       ArgusPrintRank = 0;
                                    else
                                       ArgusPrintRank++;
                                 } else
                                    if (ArgusParser->RaSOptionIndex <  ARGUS_MAX_S_OPTIONS)
                                       ArgusParser->RaSOptionStrings[ArgusParser->RaSOptionIndex++] = tok;
                                 ptr = NULL;
                              }

                              if (ArgusParser->RaSOptionIndex > 0) {
                                 ArgusProcessSOptions(ArgusParser);
                                 for (x = 0; x < ArgusParser->RaSOptionIndex; x++) 
                                    if (ArgusParser->RaSOptionStrings[x] != NULL) 
                                       ArgusParser->RaSOptionStrings[x] = NULL;
                                 RaWindowModified++;
                                 ArgusParser->RaSOptionIndex = 0;
                              }
                              RaWindowImmediate++;
                              break;
                           }
                        }

                        RaInputStatus = RANEWCOMMAND;
                        RaInputString = RANEWCOMMANDSTR;
                        RaCommandInputStr[0] = '\0';
                        break;

                     } else {
                        switch (ch) {
                           case KEY_UP: {
                              int done = 0, start = RaFilterIndex;
                              switch (RaInputStatus) {
                                 case RAGETTINGf: {
                                    do {
                                       RaFilterIndex = ((RaFilterIndex + 1) > ARGUS_DISPLAY_FILTER) ? ARGUS_REMOTE_FILTER : RaFilterIndex + 1;
                                       switch (RaFilterIndex) {
                                          case ARGUS_REMOTE_FILTER:
                                             if (ArgusParser->ArgusRemoteFilter) {
                                                sprintf (RaCommandInputStr, " remote %s", ArgusParser->ArgusRemoteFilter);
                                                RaCommandIndex = strlen(RaCommandInputStr);
                                                RaFilterIndex = ARGUS_REMOTE_FILTER;
                                                RaWindowImmediate++;
                                                done++;
                                                break;
                                             }

                                          case ARGUS_LOCAL_FILTER:
                                             if (ArgusParser->ArgusLocalFilter) {
                                                sprintf (RaCommandInputStr, " local %s", ArgusParser->ArgusLocalFilter);
                                                RaCommandIndex = strlen(RaCommandInputStr);
                                                RaFilterIndex = ARGUS_LOCAL_FILTER;
                                                RaWindowImmediate++;
                                                done++;
                                                break;
                                             }
                                          case ARGUS_DISPLAY_FILTER:
                                             if (ArgusParser->ArgusDisplayFilter) {
                                                sprintf (RaCommandInputStr, " display %s", ArgusParser->ArgusDisplayFilter);
                                                RaCommandIndex = strlen(RaCommandInputStr);
                                                RaFilterIndex = ARGUS_DISPLAY_FILTER;
                                                RaWindowImmediate++;
                                                done++;
                                                break;
                                             }
                                       }
                                    } while ((start != RaFilterIndex) && !done);
                                    break;
                                 }
                                 break;
                              }
                              break;
                           }

                           case KEY_DOWN: {
                              int trips = 0, done = 0, start = RaFilterIndex;
                              switch (RaInputStatus) {
                                 case RAGETTINGf: {
                                    do {
                                       RaFilterIndex = ((RaFilterIndex - 1) < ARGUS_REMOTE_FILTER) ? ARGUS_DISPLAY_FILTER : RaFilterIndex - 1;
                                       switch (RaFilterIndex) {
                                          case ARGUS_DISPLAY_FILTER:
                                             if (ArgusParser->ArgusDisplayFilter) {
                                                sprintf (RaCommandInputStr, " display %s", ArgusParser->ArgusDisplayFilter);
                                                RaCommandIndex = strlen(RaCommandInputStr);
                                                RaFilterIndex = ARGUS_DISPLAY_FILTER;
                                                RaWindowImmediate++;
                                                done++;
                                                break;
                                             }

                                          case ARGUS_LOCAL_FILTER:
                                             if (ArgusParser->ArgusLocalFilter) {
                                                sprintf (RaCommandInputStr, " local %s", ArgusParser->ArgusLocalFilter);
                                                RaCommandIndex = strlen(RaCommandInputStr);
                                                RaFilterIndex = ARGUS_LOCAL_FILTER;
                                                RaWindowImmediate++;
                                                done++;
                                                break;
                                             }

                                          case ARGUS_REMOTE_FILTER:
                                             if (ArgusParser->ArgusRemoteFilter) {
                                                sprintf (RaCommandInputStr, " remote %s", ArgusParser->ArgusRemoteFilter);
                                                RaCommandIndex = strlen(RaCommandInputStr);
                                                RaFilterIndex = ARGUS_REMOTE_FILTER;
                                                RaWindowImmediate++;
                                                done++;
                                                break;
                                             }
                                       }
                                       trips++;
                                    } while ((start != RaFilterIndex) && !done && (trips < 3));
                                    break;
                                 }
                                 break;
                              }
                              break;
                           }
                              
                           case KEY_LEFT:
                             if (++RaCursorOffset > RaCommandIndex)
                                RaCursorOffset = RaCommandIndex;
                             break;

                           case KEY_RIGHT:
                             if (--RaCursorOffset < 0)
                                RaCursorOffset = 0;
                             break;

                           default:
                           switch (ch) {
                              case '\b':
                              case 0x7F:
                              case KEY_DC:
                              case KEY_BACKSPACE:
                                 if (RaCursorOffset == 0) {
                                    RaCommandInputStr[RaCommandIndex--] = '\0';
                                    RaCommandInputStr[RaCommandIndex] = '\0';
                                 } else {
                                    if (RaCursorOffset < RaCommandIndex) {
                                       int z, start; 
                                       start = RaCommandIndex - (RaCursorOffset + 1);
                                       if (start < 0)
                                          start = 0;
                                       for (z = start; z < (RaCommandIndex - 1); z++)
                                          RaCommandInputStr[z] = RaCommandInputStr[z + 1];
                                       RaCommandInputStr[RaCommandIndex--] = '\0';
                                       RaCommandInputStr[RaCommandIndex] = '\0';
                                       if (RaCursorOffset > RaCommandIndex)
                                          RaCursorOffset = RaCommandIndex;
                                    }
                                 }

                                 if (RaCommandIndex < 0)
                                    RaCommandIndex = 0;
                                 break;

                              case 0x15:
                              case KEY_DL:
                                 bzero (RaCommandInputStr, MAXSTRLEN);
                                 RaCommandIndex = 0;
                                 RaCursorOffset = 0;
                                 break;
    
                              default: {
                                 if (iscntrl(ch)) {
                                    if (RaCommandInputStr[0] != '\0') {
                                       bzero (RaCommandInputStr, MAXSTRLEN);
                                       RaCommandIndex = 0;
                                       RaCursorOffset = 0;
                                    } else {
                                       bzero (RaCommandInputStr, MAXSTRLEN);
                                       RaInputStatus = RANEWCOMMAND;
                                       RaInputString = RANEWCOMMANDSTR;
                                       RaCommandIndex = 0;
                                       RaCursorOffset = 0;
                                    }

                                 } else {
                                    switch (RaInputStatus) {
                                       case RAGETTINGq:
                                          if (*RaCommandInputStr == 'y') {
                                             RaParseComplete(SIGINT);
                                          } else {
                                             RaInputStatus = RANEWCOMMAND;
                                             RaInputString = RANEWCOMMANDSTR;
                                             RaCommandInputStr[0] = '\0';
                                             RaCommandIndex = 0;
                                          }
                                          break;

                                      default:
                                         if (isascii(ch)) {
                                            if (RaCursorOffset == 0) 
                                               RaCommandInputStr[RaCommandIndex++] = ch;
                                            else {
                                               int z, start; 
                                               start = RaCommandIndex - RaCursorOffset;
                                               for (z = RaCommandIndex; z > start; z--)
                                                  RaCommandInputStr[z] = RaCommandInputStr[z-1];
                                               RaCommandInputStr[start] = ch;
                                               RaCommandIndex++;
                                            }
                                            break;
                                         }
                                    }
                                 }
                                 break;
                              }
                           }
                        }
                     }
                  }
               }
            }
         }
      }
   }

   if (RaInputStatus != RANEWCOMMAND) {
      sprintf (RaOutputBuffer, "%s%s%s", RaInputString, RaCommandInputStr, RaCommandError);
      mvwaddnstr (RaHeaderWindow, 1, 0, RaOutputBuffer, COLS - 2);
   } else
      wmove (RaHeaderWindow, 1, 0);

   getyx(RaHeaderWindow,RaCursorY,RaCursorX);
   wclrtoeol(RaHeaderWindow);

   if (RaCursesInit)
#endif
      if (ArgusParser)
         RaRefreshDisplay(ArgusParser);
}

void
parse_arg (int argc, char**argv)
{}

void
usage ()
{
   extern char version[];

#if defined(ARGUS_CURSES)
   ArgusWindowClose();
#endif

   fprintf (stderr, "Ratop Version %s\n", version);
   fprintf (stderr, "usage: %s \n", ArgusParser->ArgusProgramName);
   fprintf (stderr, "usage: %s [ra-options] [- filter-expression]\n", ArgusParser->ArgusProgramName);

   fprintf (stderr, "options: -C [host:]<port>   specify remote Cisco Netflow source.\n");
#if defined (ARGUSDEBUG)
   fprintf (stderr, "         -D <level>         specify debug level\n");
#endif
   fprintf (stderr, "         -R <directory>     recursively process argus data files in directory.\n");
   fprintf (stderr, "         -r <filename>      read argus data filename.\n");
   fprintf (stderr, "         -S <host[:port]>   specify remote argus <host> and optional port\n");
   fprintf (stderr, "                            number.\n");
#if defined(ARGUS_SASL)
   fprintf (stderr, "         -U <user/auth>     specify <user/auth> authentication information.\n");
#endif
   exit(1);
}


/*
   RaProcessRecord - this routine will take a non-managment record and
   process it as if it were a SBP.  This basically means, transform the
   flow descriptor to whatever model is appropriate, find the flow
   cache.  Then we carve the new record into the appropriate size for
   the SBP operation, and then proceed to merge the fragments into the
   appropriate record for this ns.

   If the ns cache is a sticky ns, it may not be in the RaTopProcess
   queue, so we need to check and put it in if necessary.

   And because we had a record, we'll indicate that the window needs
   to be updated.

   All screen operations, queue timeouts etc, are done in 
   ArgusClientTimeout, so we're done here.

*/
void RaProcessRecord (struct ArgusParserStruct *, struct ArgusRecordStruct *);
void RaProcessThisRecord (struct ArgusParserStruct *, struct ArgusRecordStruct *);

void
RaProcessRecord (struct ArgusParserStruct *parser, struct ArgusRecordStruct *ns)
{
   if (ns->hdr.type & ARGUS_MAR) {
      RaProcessManRecord(parser, ns);
   } else {
      if (parser->RaMonMode) {
         struct ArgusRecordStruct *tns = ArgusCopyRecordStruct(ns);
         RaProcessThisRecord(parser, ns);
         ArgusReverseRecord(tns);
         RaProcessThisRecord(parser, tns);
         ArgusDeleteRecordStruct(parser, tns);

      } else {
         if (ArgusAggregator->ArgusMatrixMode) {
            struct ArgusFlow *flow = (struct ArgusFlow *) &ns->canon.flow;

            switch (flow->hdr.subtype & 0x3F) {
               case ARGUS_FLOW_LAYER_3_MATRIX:
               case ARGUS_FLOW_CLASSIC5TUPLE: {
                  switch (flow->hdr.argus_dsrvl8.qual & 0x1F) {
                     case ARGUS_TYPE_IPV4: {
                        if (flow->ip_flow.ip_src > flow->ip_flow.ip_dst)
                           ArgusReverseRecord(ns);
                     }
                     break;

                     case ARGUS_TYPE_IPV6: {
                        int i;
                        for (i = 0; i < 4; i++) {
                           if (flow->ipv6_flow.ip_src[i] < flow->ipv6_flow.ip_dst[i])
                              break;

                           if (flow->ipv6_flow.ip_src[i] > flow->ipv6_flow.ip_dst[i]) {
                              ArgusReverseRecord(ns);
                              break;
                           }
                        }
                     }
                     break;
                  }
                  break;
               }

               default:
                  break;
            }
         }

         RaProcessThisRecord(parser, ns);
      }
   }
}

void
RaProcessThisRecord (struct ArgusParserStruct *parser, struct ArgusRecordStruct *ns)
{
   struct ArgusRecordStruct *tns = NULL, *pns = NULL;
   struct ArgusRecordStruct *cns = ArgusCopyRecordStruct(ns);
   struct ArgusAggregatorStruct *agg = ArgusAggregator;
   struct ArgusHashStruct *hstruct = NULL;
   int found = 0;

   if (ns->hdr.type & ARGUS_MAR) {
      RaProcessManRecord(parser, ns);

   } else {
      ArgusThisTime.tv_sec  = ns->canon.time.src.start.tv_sec;
      ArgusThisTime.tv_usec = ns->canon.time.src.start.tv_usec;

      if (ArgusLastTime.tv_sec == 0) {
         ArgusLastTime    = ArgusThisTime;
         ArgusCurrentTime = ArgusThisTime;
      }

      if (!((ArgusLastTime.tv_sec  > ArgusThisTime.tv_sec) ||
         ((ArgusLastTime.tv_sec == ArgusThisTime.tv_sec) &&
          (ArgusLastTime.tv_usec > ArgusThisTime.tv_usec)))) {

         while (!(RaUpdateRate)) {
            usleep (5000);
            ArgusClientTimeout ();
         }

/* ok so lets deal with realtime processing */
         if (!(parser->Sflag) && (RaTopRealTime)) {
            if ((ArgusThisTime.tv_sec  > ArgusLastTime.tv_sec) ||
               ((ArgusThisTime.tv_sec == ArgusLastTime.tv_sec) &&
                (ArgusThisTime.tv_usec > ArgusLastTime.tv_usec))) {
               int thisRate;
               int deltausec;

/* this record is some period of time after the last record, so 
   lets calculate the difference, and then sleep to deal with
   time that needs to lapse */

               dRealTime = *RaDiffTime(&ArgusThisTime, &ArgusLastTime);
               thisUsec  = ((dRealTime.tv_sec * 1000000) + dRealTime.tv_usec)/RaUpdateRate;

               dRealTime = *RaDiffTime(&parser->ArgusRealTime, &ArgusLastRealTime);
               lastUsec  = ((dRealTime.tv_sec * 1000000) + dRealTime.tv_usec)/RaUpdateRate;

               while ((deltausec = (thisUsec - lastUsec)) > 0) {
                  thisRate = (deltausec > 50000) ? 50000 : deltausec;
#if defined(ARGUSDEBUG)
                  ArgusDebug (6, "ArgusProcessThisRecord () idling needed for %d usecs\n", deltausec); 
#endif
                  usleep (thisRate);
                  ArgusClientTimeout ();

                  dRealTime = *RaDiffTime(&parser->ArgusRealTime, &ArgusLastRealTime);
                  lastUsec  = ((dRealTime.tv_sec * 1000000) + dRealTime.tv_usec)/RaUpdateRate;
               }
            }
         }

         ArgusLastRealTime = parser->ArgusRealTime;
         ArgusLastTime     = ArgusThisTime;
         ArgusCurrentTime  = ArgusThisTime;
      }

      RaTopStopTime = parser->ArgusRealTime;
      if (ArgusParser->RaClientUpdate.tv_sec == 0) {
         ArgusParser->RaClientUpdate.tv_sec = parser->ArgusGlobalTime.tv_sec;
         ArgusParser->RaClientUpdate.tv_usec = 0;
      }
      if (RaTopStartTime.tv_sec == 0)
         RaTopStartTime = parser->ArgusRealTime;

      while (agg && !found) {
         struct nff_insn *fcode = agg->filter.bf_insns;

         if (ArgusFilterRecord (fcode, ns) != 0) {
            if ((agg->rap = RaFlowModelOverRides(agg, cns)) == NULL)
               agg->rap = agg->drap;

            ArgusGenerateNewFlow(agg, cns);

            if ((hstruct = ArgusGenerateHashStruct(agg, cns, &ns->canon.flow)) == NULL)
               ArgusLog (LOG_ERR, "RaProcessRecord: ArgusGenerateHashStruct error %s", strerror(errno));

            if ((pns = ArgusFindRecord(RaTopProcess->htable, hstruct)) == NULL) {
               struct ArgusFlow *flow = (struct ArgusFlow *) &cns->canon.flow;
               int tryreverse = 1;

               switch (flow->hdr.argus_dsrvl8.qual & 0x1F) {
                  case ARGUS_TYPE_IPV4: {
                     switch (flow->ip_flow.ip_p) {
                        case IPPROTO_ESP:
                           tryreverse = 0;
                           break;
                     }
                  }
               }

               if (!parser->RaMonMode && tryreverse) {
                  struct ArgusRecordStruct *dns = ArgusCopyRecordStruct(ns);

                  ArgusReverseRecord (dns);
                  ArgusGenerateNewFlow(agg, dns);

                  if ((hstruct = ArgusGenerateHashStruct(agg, dns, &ns->canon.flow)) == NULL)
                     ArgusLog (LOG_ERR, "RaProcessThisRecord: ArgusGenerateHashStruct error %s", strerror(errno));

                  if ((pns = ArgusFindRecord(RaTopProcess->htable, hstruct)) != NULL) {
                     ArgusDeleteRecordStruct(ArgusParser, cns);
                     cns = dns;

                  } else {
                     ArgusDeleteRecordStruct(ArgusParser, dns);
                     if ((hstruct = ArgusGenerateHashStruct(agg, cns, &ns->canon.flow)) == NULL)
                        ArgusLog (LOG_ERR, "RaProcessThisRecord: ArgusGenerateHashStruct error %s", strerror(errno));
                  }
               }
            }

            if ((pns) && pns->qhdr.queue) {
                  ArgusRemoveFromQueue (pns->qhdr.queue, &pns->qhdr);
               ArgusAddToQueue (RaTopProcess->queue, &pns->qhdr);
               pns->status |= ARGUS_RECORD_MODIFIED;
            }
            found++;

         } else
            agg = agg->nxt;
      }

      if (!found)
         if ((hstruct = ArgusGenerateHashStruct(agg, cns, &ns->canon.flow)) == NULL)
            ArgusLog (LOG_ERR, "RaProcessThisRecord: ArgusGenerateHashStruct error %s", strerror(errno));

      RaBinProcess->nadp.tperiod = 0.0;

      while ((tns = ArgusAlignRecord(parser, cns, &RaBinProcess->nadp)) != NULL) {
         int offset = 0;

         if (pns) {
            if (pns->bins) {
               offset = parser->Bflag / pns->bins->size;
               pns->bins->nadp.RaStartTmStruct = RaBinProcess->nadp.RaStartTmStruct;
               pns->bins->nadp.RaEndTmStruct   = RaBinProcess->nadp.RaEndTmStruct;

               if (!(ArgusInsertRecord (parser, pns->bins, tns, offset) > 0))
                  ArgusDeleteRecordStruct(ArgusParser, tns);

               pns->bins->status |= RA_DIRTYBINS;

            } else {
               if (parser->RaCumulativeMerge) {
                  ArgusMergeRecords (ArgusAggregator, pns, tns);
                  ArgusDeleteRecordStruct(ArgusParser, tns);

               } else 
                  bcopy ((char *)&tns->canon, (char *)&pns->canon, sizeof(pns->canon));

               pns->status |= ARGUS_RECORD_MODIFIED;
            }

            ArgusRemoveFromQueue(RaTopProcess->queue, &pns->qhdr);
            ArgusAddToQueue (RaTopProcess->queue, &pns->qhdr);

         } else {
            if ((pns =  ArgusCopyRecordStruct(tns)) != NULL) { /* new record */
               if (RaBinProcess->nadp.mode == ARGUSSPLITRATE) {
                  if ((pns->bins = (struct RaBinProcessStruct *)ArgusNewRateBins(parser, pns)) == NULL)
                     ArgusLog (LOG_ERR, "ArgusProcessThisRecord: ArgusNewRateBins error %s", strerror(errno));

                  offset = parser->Bflag / pns->bins->size;

                  if (!(ArgusInsertRecord (parser, pns->bins, tns, offset) > 0)) 
                     ArgusDeleteRecordStruct(ArgusParser, tns);

                  pns->bins->status |= RA_DIRTYBINS;

               } else
                  ArgusDeleteRecordStruct(ArgusParser, tns);

               pns->status |= ARGUS_RECORD_MODIFIED;

               pns->htblhdr = ArgusAddHashEntry (RaTopProcess->htable, pns, hstruct);
               ArgusAddToQueue (RaTopProcess->queue, &pns->qhdr);
            }
         }
      }

      RaWindowModified = 1;
   }

   ArgusDeleteRecordStruct(ArgusParser, cns);

#if defined(ARGUSDEBUG)
   ArgusDebug (6, "ArgusProcessThisRecord () returning\n"); 
#endif
}

void
RaProcessManRecord (struct ArgusParserStruct *parser, struct ArgusRecordStruct *ns)
{
 
#if defined(ARGUSDEBUG)
   ArgusDebug (6, "ArgusProcessManRecord () returning\n"); 
#endif
}

int RaSendArgusRecord(struct ArgusRecordStruct *ns) {return 0;}

void ArgusWindowClose(void);
int ArgusWindowClosing = 0;
void
ArgusWindowClose(void)
{ 
   if (!(ArgusWindowClosing++)) {
#if defined(ARGUS_CURSES)
      if (RaCursesInit && (!(isendwin()))) {
         while (wgetch(RaWindow) != ERR) ;
         endwin();
         printf("\n");
      }
#endif
   }

#if defined(ARGUSDEBUG)
   ArgusDebug (6, "ArgusWindowClose () returning\n"); 
#endif
}


#if defined(ARGUS_CURSES)
int 
RaInitCurses (struct ArgusParserStruct *parser)
{
   RaCursesInit++;

   parser->RaCursesMode = 1;
   RaWindow = initscr();

   noraw();
   cbreak();
   keypad(stdscr, TRUE);
   meta(stdscr, TRUE);
   nodelay(RaWindow, TRUE);
   noecho();
   nonl();
   intrflush(stdscr, FALSE);

   clearok(RaWindow, TRUE);
   werase(RaWindow);
   wrefresh(RaWindow);
 
   RaHeaderWindow = newwin (RaHeaderWinSize, COLS, 0, 0);
   RaWindowLines  = LINES - (RaHeaderWinSize + 1);
   RaWindowStartLine = 0;
   RaDisplayLines = RaWindowLines - 2;

   RaAvailableWindow = newwin (RaWindowLines, COLS, RaHeaderWinSize, 0);
   RaFailedWindow = RaAvailableWindow;

   idlok (RaHeaderWindow, TRUE);
   idlok (RaAvailableWindow, TRUE);
   notimeout(RaHeaderWindow, TRUE);
   refresh();
   return (1);
}


void
RaResizeScreen(void)
{
   RaScreenLines = LINES;
   RaScreenColumns = COLS;

   RaWindowLines = LINES - (RaHeaderWinSize + 1);
   RaDisplayLines = RaWindowLines - 2;

   werase(RaWindow);
 
#if !defined(HAVE_SOLARIS)
   wresize(RaWindow, LINES, COLS);
   wresize(RaHeaderWindow, RaHeaderWinSize, COLS);
#else
   delwin(RaHeaderWindow);
   RaHeaderWindow = newwin (RaHeaderWinSize, COLS, 0, 0);
   idlok (RaHeaderWindow, TRUE);
   notimeout(RaHeaderWindow, TRUE);
#endif /* HAVE_SOLARIS */
 
   RaAvailableWindow = newwin (RaWindowLines, COLS, RaHeaderWinSize, 0);
   RaFailedWindow = RaAvailableWindow;
   idlok (RaAvailableWindow, TRUE);
   ArgusParser->RaLabel = NULL;
   RaWindowModified = RA_MODIFIED;
   RaWindowImmediate = TRUE;
   RaScreenResize = FALSE;
}


void
RaOutputModifyScreen ()
{
   int i = 0;
   werase(RaAvailableWindow);
 
   for (i = RaMinCommandLines; i < (RaMaxCommandLines + 1); i++) {
      mvwprintw (RaAvailableWindow, i, 1, RaCommandArray[i - RaMinCommandLines]);
      if (i == RaMinCommandLines)
         wstandout(RaAvailableWindow);
      wprintw (RaAvailableWindow, "%s", RaCommandValueArray[i - RaMinCommandLines]());
      if (i == RaMinCommandLines)
         wstandend(RaAvailableWindow);
   }
}

void
RaOutputHelpScreen ()
{
   extern char version[];
   werase(RaAvailableWindow);
   mvwprintw (RaAvailableWindow, 1, 1, "RaTop Version %s\n", version);
   mvwprintw (RaAvailableWindow, 2, 1, "Key Commands: c,d,D,f,F,h,m,n,N,p,P,q,r,R,s,S,t,T,u,U,v,w,z,Z,=");
   mvwprintw (RaAvailableWindow, 4, 1, "  ^D - Clear command line. Reset input (also ESC).");
   mvwprintw (RaAvailableWindow, 5, 1, "   c - Connect to remote Argus Source");
   mvwprintw (RaAvailableWindow, 6, 1, "   d - Drop connection from remote argus source");
   mvwprintw (RaAvailableWindow, 7, 1, "   D - Set debug printing level");
   mvwprintw (RaAvailableWindow, 8, 1, "   f - Specify filter expression");
   mvwprintw (RaAvailableWindow, 9, 1, "   F - Specify fields to print (use arrow keys to navigate).");
   mvwprintw (RaAvailableWindow,10, 1, "         +[#]field - add field to optional column # or end of line");
   mvwprintw (RaAvailableWindow,11, 1, "         -field    - remove field from display");
   mvwprintw (RaAvailableWindow,12, 1, "          field    - reset fields and add to display");
   mvwprintw (RaAvailableWindow,13, 1, "             available fields are:");
   mvwprintw (RaAvailableWindow,14, 1, "               srcid, stime, ltime, dur, avgdur, trans, flgs, dir, state, seq, bins, binnum");
   mvwprintw (RaAvailableWindow,15, 1, "               mac, smac, dmac, mpls, smpls, dmpls, vlan, svlan, dvlan, svid, dvid, svpri, dvpri");
   mvwprintw (RaAvailableWindow,16, 1, "               saddr, daddr, snet, dnet, proto, sport, dport, stos, dtos, sttl, dttl, sipid, dipid");
   mvwprintw (RaAvailableWindow,17, 1, "               tcpext, tcprtt, stcpb, dtcpb, swin, dwin, srng, drng, spksz, dpksz, smaxsz, sminsz, dmaxsz, dminsz");
   mvwprintw (RaAvailableWindow,18, 1, "               suser, duser, svc, pkts, spkts, dpkts, load,sload, dload, bytes, sbytes, dbytes, rate, srate, drate");
   mvwprintw (RaAvailableWindow,19, 1, "               sloss, dloss, sintpkt, dintpkt, sjit, djit, sintpktact, dintpktact, sintpktidl, dintpktidl");
   mvwprintw (RaAvailableWindow,20, 1, "               sjitidl, djitidl, ddur, dstime, dltime, dspkts, ddpkts, dsbytes, ddbytes");
   mvwprintw (RaAvailableWindow,21, 1, "               djitact, jitidl, sjitidl, djitidl, state, ddur, dstime, dltime, dspkts, ddpkts");
   mvwprintw (RaAvailableWindow,22, 1, "   m - Specify the flow model objects.");
   mvwprintw (RaAvailableWindow,23, 1, "   n - Toggle name to number conversion(cycle through).");
   mvwprintw (RaAvailableWindow,24, 1, "   N - Specify the number of items to print.");
   mvwprintw (RaAvailableWindow,25, 1, "   p - Toggle metric percentage reporting.");
   mvwprintw (RaAvailableWindow,26, 1, "   P - Pause the program");
   mvwprintw (RaAvailableWindow,27, 1, "   q - Quit the program.");
   mvwprintw (RaAvailableWindow,28, 1, "   r - Read argus data file(s)");
   mvwprintw (RaAvailableWindow,29, 1, "   R - Recursively open argus data files(s)");
   mvwprintw (RaAvailableWindow,30, 1, "   s - Specify sort fields.");
   mvwprintw (RaAvailableWindow,31, 1, "   t - Specify time range. same as -t command line option. ");
   mvwprintw (RaAvailableWindow,32, 1, "   T - Specify idle timeout value [60s].");
   mvwprintw (RaAvailableWindow,33, 1, "   u - Specify the window update timer, in seconds [0.1s]");
   mvwprintw (RaAvailableWindow,34, 1, "   U - Specify the playback rate, in seconds per second [1.0]");
   mvwprintw (RaAvailableWindow,35, 1, "   v - reverse the sort order");
   mvwprintw (RaAvailableWindow,36, 1, "   w - Write display to file");
   mvwprintw (RaAvailableWindow,37, 1, "   z - Toggle State field output formats");
   mvwprintw (RaAvailableWindow,38, 1, "   Z - Toggle TCP State field output");
   mvwprintw (RaAvailableWindow,39, 1, "   = - Clear Flow List");
   mvwprintw (RaAvailableWindow,40, 1, "   h - Print help screen.");
   mvwprintw (RaAvailableWindow,42, 1, "Navigation Keys (vi): g,G,j,k,i^F,^D,^B,^U");

   wnoutrefresh(RaAvailableWindow);
   doupdate();
}


#endif /* ARGUS_CURSES */


struct RaTopProcessStruct *
RaTopNewProcess(struct ArgusParserStruct *parser)
{
   struct RaTopProcessStruct *retn = NULL;
 
   if ((retn = (struct RaTopProcessStruct *) ArgusCalloc (1, sizeof(*retn))) != NULL) {
      if ((retn->queue = ArgusNewQueue()) == NULL)
         ArgusLog (LOG_ERR, "RaTopNewProcess: ArgusNewQueue error %s\n", strerror(errno));

      if ((retn->delqueue = ArgusNewQueue()) == NULL)
         ArgusLog (LOG_ERR, "RaTopNewProcess: ArgusNewQueue error %s\n", strerror(errno));
 
      if ((retn->htable = ArgusNewHashTable(0x100000)) == NULL)
         ArgusLog (LOG_ERR, "RaTopNewProcess: ArgusCalloc error %s\n", strerror(errno));

   } else
      ArgusLog (LOG_ERR, "RaTopNewProcess: ArgusCalloc error %s\n", strerror(errno));
 
#ifdef ARGUSDEBUG
   ArgusDebug (3, "RaTopNewProcess(0x%x) returns 0x%x\n", parser, retn);
#endif
   return (retn);
}



char RaGetStrBuf[MAXSTRLEN];

char *
RaGetCiscoServers(void)
{
   char *retn = RaGetStrBuf;

   bzero(RaGetStrBuf, MAXSTRLEN);
   sprintf(RaGetStrBuf, "%s", (ArgusParser->Cflag ? "yes" : "no"));
   return(retn);
}

char *
RaGetNoOutputStatus(void)
{
   char *retn = RaGetStrBuf;

   bzero(RaGetStrBuf, MAXSTRLEN);
   sprintf(RaGetStrBuf, "%s", (ArgusParser->qflag ? "yes" : "no"));
   return(retn);
}

char *
RaGetUserAuth(void)
{
   char *retn = RaGetStrBuf;

   bzero(RaGetStrBuf, MAXSTRLEN);
   sprintf(RaGetStrBuf, "%s", ArgusParser->ustr);
   return(retn);
}

char *
RaGetUserPass(void)
{
   char *retn = RaGetStrBuf;
   int i;

   bzero(RaGetStrBuf, MAXSTRLEN);
   sprintf(RaGetStrBuf, "%s", ArgusParser->pstr);
   for (i = 0; i < strlen(RaGetStrBuf); i++)
      RaGetStrBuf[i] = 'x';
   return(retn);
}

char *
RaGetOutputFile(void)
{
   char *retn = RaGetStrBuf;
   struct ArgusWfileStruct *wfile = NULL, *start;

   bzero(RaGetStrBuf, MAXSTRLEN);

   if (ArgusParser->ArgusWfileList != NULL) {
      if ((wfile = (struct ArgusWfileStruct *) ArgusFrontList(ArgusParser->ArgusWfileList)) != NULL) {
         start = wfile;
         do {
            sprintf(&RaGetStrBuf[strlen(RaGetStrBuf)], "%s ", wfile->filename);
            ArgusPopFrontList(ArgusParser->ArgusWfileList, ARGUS_NOLOCK);
            ArgusPushBackList(ArgusParser->ArgusWfileList, (struct ArgusListRecord *)wfile, ARGUS_NOLOCK);
            wfile = (struct ArgusWfileStruct *)ArgusFrontList(ArgusParser->ArgusWfileList);
         } while (wfile != start);
      }
      sprintf(RaGetStrBuf, "%s", RaGetStrBuf);
   }
   return(retn);
}

char *
RaGetExceptionOutputFile(void)
{
   char *retn = RaGetStrBuf;

   bzero(RaGetStrBuf, MAXSTRLEN);
   sprintf(RaGetStrBuf, "%s", ArgusParser->exceptfile);
   return(retn);
}

char *
RaGetTimeRange(void)
{
   char *retn = RaGetStrBuf;

   bzero(RaGetStrBuf, MAXSTRLEN);
   sprintf(RaGetStrBuf, "%s", ArgusParser->timearg);
   return(retn);
}

char *
RaGetRunTime(void)
{
   char *retn = RaGetStrBuf;

   bzero(RaGetStrBuf, MAXSTRLEN);
   sprintf(RaGetStrBuf, "%d", ArgusParser->Tflag);
   return(retn);
}

char *
RaGetFieldDelimiter(void)
{
   char *retn = RaGetStrBuf;

   bzero(RaGetStrBuf, MAXSTRLEN);
   if (ArgusParser->RaFieldDelimiter == '\0')
      sprintf(RaGetStrBuf, "'\\0'");
   else
      sprintf(RaGetStrBuf, "'%c'", ArgusParser->RaFieldDelimiter);
   return(retn);
}

char *
RaGetTimeFormat(void)
{
   char *retn = RaGetStrBuf;

   bzero(RaGetStrBuf, MAXSTRLEN);
   sprintf(RaGetStrBuf, "%s", ArgusParser->RaTimeFormat);
   return(retn);
}

char *
RaGetPrecision(void)
{
   char *retn = RaGetStrBuf;

   bzero(RaGetStrBuf, MAXSTRLEN);
   sprintf(RaGetStrBuf, "%d", ArgusParser->pflag);
   return(retn);
}

char *
RaGetTimeSeries(void)
{
   char *retn = RaGetStrBuf;

   bzero(RaGetStrBuf, MAXSTRLEN);
   sprintf(RaGetStrBuf, "%s", (ArgusParser->Hstr ? "yes" : "no"));
   return(retn);
}

char *
RaGetValidateStatus(void)
{
   char *retn = RaGetStrBuf;

   bzero(RaGetStrBuf, MAXSTRLEN);
   sprintf(RaGetStrBuf, "%s", (ArgusParser->Vflag ? "yes" : "no"));
   return(retn);
}

char *
RaGetNumber(void)
{
   char *retn = RaGetStrBuf;

   bzero(RaGetStrBuf, MAXSTRLEN);
   sprintf(RaGetStrBuf, "%d", ArgusParser->eNflag);
   return(retn);
}

char *
RaGetDebugLevel(void)
{
   char *retn = RaGetStrBuf;

   bzero(RaGetStrBuf, MAXSTRLEN);
   sprintf(RaGetStrBuf, "%d", ArgusParser->debugflag);
   return(retn);
}

char *
RaGetUserDataEncode(void)
{
   char *retn = RaGetStrBuf, *str = NULL;

   bzero(RaGetStrBuf, MAXSTRLEN);
   switch (ArgusParser->eflag) {
      case ARGUS_ENCODE_ASCII:
         str = "ascii"; break;
      case ARGUS_ENCODE_32:
         str = "encode32"; break;
      case ARGUS_ENCODE_64:
         str = "encode64"; break;
   }

   sprintf(RaGetStrBuf, "%s", str);
   return(retn);
}

char *
RaGetHostFieldLength(void)
{
   char *retn = RaGetStrBuf;

   bzero(RaGetStrBuf, MAXSTRLEN);
   sprintf(RaGetStrBuf, "%d", ArgusParser->hfield);
   return(retn);
}

char *
RaGetPortFieldLength(void)
{
   char *retn = RaGetStrBuf;

   bzero(RaGetStrBuf, MAXSTRLEN);
   sprintf(RaGetStrBuf, "%d", ArgusParser->pfield);
   return(retn);
}



char *
RaSplitFilename (struct ArgusAdjustStruct *nadp)
{
   char *retn = NULL, tmpbuf[MAXSTRLEN];
   char *filename = nadp->filename;
   int len, i = 1, carry = 0;

   if (filename != NULL) {
      len = strlen(filename);

      for (i = 0; i < nadp->slen; i++)
         if (filename[len - (i + 1)] == 'z')
            carry++;

      if ((carry == (nadp->slen - 1)) && (filename[len - nadp->slen] == 'y')) {
         bzero (tmpbuf, MAXSTRLEN);
         strncpy(tmpbuf, filename, MAXSTRLEN);
         tmpbuf[strlen(tmpbuf) - nadp->slen] = 'z';
         for (i = 0; i < nadp->slen; i++)
            strcat(tmpbuf, "a");
         nadp->slen++;

      } else {
         for (i = 0, carry = 0; i < nadp->slen; i++) {
            if (filename[len - (i + 1)] == 'z') {
               filename[len - (i + 1)] = 'a';
            } else {
               filename[len - (i + 1)]++;
               break;
            }
         }
         strncpy (tmpbuf, filename, MAXSTRLEN);
      }

      if (nadp->filename)
         free(nadp->filename);

      nadp->filename = strdup(tmpbuf);
      retn = nadp->filename;
   }


#ifdef ARGUSDEBUG
   ArgusDebug (5, "RaSplitFilename (0x%x) returning %s\n", nadp, retn); 
#endif

   return (retn);
}

int
RaProcessSplitOptions(struct ArgusParserStruct *parser, char *str, int len, struct ArgusRecordStruct *ns)
{
   char resultbuf[MAXSTRLEN], tmpbuf[MAXSTRLEN];
   char *ptr = NULL, *tptr = str;
   int retn = 0, i, x;

   bzero (resultbuf, len);

   while ((ptr = strchr (tptr, '$')) != NULL) {
      *ptr++ = '\0';
      sprintf (&resultbuf[strlen(resultbuf)], tptr);

      for (i = 0, x = 0; x < MAX_PRINT_ALG_TYPES; x++) {
         if (!strncmp (RaPrintAlgorithmTable[x].field, ptr, strlen(RaPrintAlgorithmTable[x].field))) {
            bzero (tmpbuf, MAXSTRLEN);
            RaPrintAlgorithmTable[x].print(parser, tmpbuf, ns);

            while (isspace((int)tmpbuf[strlen(tmpbuf) - 1]))
               tmpbuf[strlen(tmpbuf) - 1] = '\0';

            while (isspace((int)tmpbuf[i])) i++;
            sprintf (&resultbuf[strlen(resultbuf)], "%s", &tmpbuf[i]);

            ptr += strlen(RaPrintAlgorithmTable[x].field);
            while (*ptr && (*ptr != '$'))
               bcopy (ptr++, &resultbuf[strlen(resultbuf)], 1);
            break;
         }
      }

      tptr = ptr;
      retn++;
   }

   if (retn) {
      bzero (str, len);
      bcopy (resultbuf, str, strlen(resultbuf));
   }

#ifdef ARGUSDEBUG
   ArgusDebug (1, "RaProcessSplitOptions(%s, %d, 0x%x): returns %d", str, len, ns, retn);
#endif

   return (retn);
}


